import pydantic
import pytest

from compass.core.schemas import hierarchy

base_data = {
    "unit_id": 42,
    "name": "name",
}


class TestSchemaHierarchy:
    def test_hierarchy_base_round_trip(self):
        # Given
        data = dict(unit_id=123)

        # When
        result = hierarchy.HierarchyBase(**data)

        # Then
        assert data == result.dict()

    def test_hierarchy_base_invalid(self):
        # Given
        data = dict(unit_id="abc")

        # Then
        with pytest.raises(pydantic.ValidationError, match=r"validation errors? for HierarchyBase"):
            # When
            hierarchy.HierarchyBase(**data)

    def test_hierarchy_unit_round_trip(self):
        # Given
        data = base_data.copy()

        # When
        result = hierarchy.HierarchyUnit(**data)

        # Then
        assert data == result.dict()

    def test_hierarchy_unit_invalid_normal(self):
        # Given
        data = base_data | dict(name={"W1A 1AA", "SW1A 1AA"})  # type: ignore # testing

        # Then
        with pytest.raises(pydantic.ValidationError, match=r"validation errors? for HierarchyUnit"):
            # When
            hierarchy.HierarchyUnit(**data)

    def test_hierarchy_section_round_trip(self):
        # Given
        data = base_data | dict(section_type="Beavers")

        # When
        result = hierarchy.HierarchySection(**data)

        # Then
        assert data == result.dict()

    def test_hierarchy_section_invalid_literal(self):
        # Given
        data = base_data | dict(section_type="random")

        # Then
        with pytest.raises(pydantic.ValidationError, match=r"validation errors? for HierarchySection"):
            # When
            hierarchy.HierarchySection(**data)

    def test_hierarchy_level_round_trip(self):
        # Given
        data = dict(unit_id=123, level="Group")

        # When
        result = hierarchy.HierarchyLevel(**data)

        # Then
        assert data == result.dict()

    def test_hierarchy_level_invalid_literal(self):
        # Given
        data = dict(unit_id=123, level="random")

        # Then
        with pytest.raises(pydantic.ValidationError, match=r"validation errors? for HierarchyLevel"):
            # When
            hierarchy.HierarchyLevel(**data)

    def test_unit_data_round_trip(self):
        # Given
        section_data = base_data | dict(unit_id=1, section_type="Beavers")
        data = dict(unit_id=42, level="Group", child=None, sections=[section_data])

        # When
        result = hierarchy.UnitData(**data)

        # Then
        assert data | {"name": None} == result.dict()

    def test_unit_data_invalid_missing(self):
        # Given
        data = dict(unit_id=123, level="Group", child=None)

        # Then
        with pytest.raises(pydantic.ValidationError, match=r"validation errors? for UnitData"):
            # When
            hierarchy.UnitData(**data)

    def test_descendant_data_round_trip(self):
        # Given
        hierarchy_unit_data = dict(unit_id=123, name="name")
        unit_data_data = dict(unit_id=123, level="Group", child=None, sections=[hierarchy_unit_data | dict(section_type="Beavers")])
        data = hierarchy_unit_data | unit_data_data

        # When
        result = hierarchy.DescendantData(**data)

        # Then
        assert data == result.dict()

    def test_descendant_data_invalid_missing(self):
        # Given
        hierarchy_unit_data = dict(id=123, name="name")
        data = hierarchy_unit_data | {}

        # Then
        with pytest.raises(pydantic.ValidationError, match=r"validation errors? for DescendantData"):
            # When
            hierarchy.DescendantData(**data)

    def test_hierarchy_member_round_trip(self):
        # Given
        data = dict(contact_number=123, name="Adam Smith", role=None)

        # When
        result = hierarchy.HierarchyMember(**data)

        # Then
        assert data == result.dict()

    def test_hierarchy_member_invalid(self):
        # Given
        data = dict(contact_number=123, name="Adam Smith", role=object())

        # Then
        with pytest.raises(pydantic.ValidationError, match=r"validation errors? for HierarchyMember"):
            # When
            hierarchy.HierarchyMember(**data)

    def test_hierarchy_unit_members_round_trip(self):
        # Given
        data = dict(unit_id=321, members=[dict(contact_number=123, name="Adam Smith", role=None)])

        # When
        result = hierarchy.HierarchyUnitMembers(**data)

        # Then
        assert data == result.dict()

    def test_hierarchy_unit_members_invalid_missing(self):
        # Given
        data = dict(unit_id=321, members=[object()])

        # Then
        with pytest.raises(pydantic.ValidationError, match=r"validation errors? for HierarchyUnitMembers"):
            # When
            hierarchy.HierarchyUnitMembers(**data)

    def test_hierarchy_unit_members_list_round_trip(self):
        # Given
        data = [dict(unit_id=321, members=[dict(contact_number=123, name="Adam Smith", role=None)])]

        # When
        result = [hierarchy.HierarchyUnitMembers(**unit_members) for unit_members in data]

        # Then
        assert data == [model.__dict__ for model in result]



import pytest

from compass.core.schemas import member

base_data = {
    "membership_number": 0,
    "name": "...",
    "known_as": "...",
}


class TestSchemaMember:
    def test_member_details_phone_number_valid(self):
        # Given
        number_landline = "01212345679"  # Fictitious number
        number_mobile = "07707123456"  # Fictitious number
        number_international = "+1 2015550123"  # Fictitious international number
        # https://www.ofcom.org.uk/phones-telecoms-and-internet/information-for-industry/numbering/numbers-for-drama

        #  When
        result_landline = member.MemberDetails(main_phone=number_landline, **base_data)
        result_mobile = member.MemberDetails(main_phone=number_mobile, **base_data)
        result_international = member.MemberDetails(main_phone=number_international, **base_data)

        # Them
        assert result_landline.main_phone == "0121 234 5679"
        assert result_mobile.main_phone == "07707 123456"
        assert result_international.main_phone == "+1 201-555-0123"

    def test_member_details_phone_number_passing(self):
        # Given
        number_none = None  # Literal `None`
        number_zero = "0"  # Literal string `0`
        number_len_1 = "7"  # Strings of length one
        number_len_0 = ""  # Empty strings

        #  When
        result_none = member.MemberDetails(main_phone=number_none, **base_data)
        result_zero = member.MemberDetails(main_phone=number_zero, **base_data)
        result_len_1 = member.MemberDetails(main_phone=number_len_1, **base_data)
        result_len_0 = member.MemberDetails(main_phone=number_len_0, **base_data)

        # Them
        assert None is result_none.main_phone
        assert None is result_zero.main_phone
        assert None is result_len_1.main_phone
        assert None is result_len_0.main_phone

    def test_member_details_phone_number_warning(self):
        # Given
        number_invalid_1 = "abc"  # Fictitious invalid number
        number_invalid_2 = "07700 90042"  # Fictitious invalid number
        # https://www.ofcom.org.uk/phones-telecoms-and-internet/information-for-industry/numbering/numbers-for-drama

        # Them
        with pytest.warns(RuntimeWarning, match=f"Member No 0: phone number {number_invalid_1} is not valid!"):
            #  When
            member.MemberDetails(main_phone=number_invalid_1, **base_data)

        # Them
        with pytest.warns(RuntimeWarning, match=f"Member No 0: phone number {number_invalid_2} is not valid!"):
            #  When
            member.MemberDetails(main_phone=number_invalid_2, **base_data)



import datetime

from compass.core._scrapers import contact


class TestContact:
    def test_process_address(self):
        # Given
        addresses = [
            {
                "Line1": "Registry",
                "Line2": "Old College",
                "Line3": "City of Edinburgh",
                "Town": "EDINBURGH",
                "County": "Midlothian",
                "Postcode": "EH1 1AA",
                "Country": "UK",
                "ValidFrom": None,
                "IsMain": True,
                "Historical": False,
                "AddressNumber": 123456,
                "Address": "Registry, Old College, City of Edinburgh, EDINBURGH, Midlothian. EH1 1AA UK",
            }
        ]

        # When
        result = contact._process_address(addresses)

        # Then
        assert isinstance(result, dict)
        assert all(key.__class__ is str for key in result.keys())
        assert all(value.__class__ is str for value in result.values())
        assert [*result.keys()] == ["unparsed_address", "country", "postcode", "county", "town", "street"]
        assert result == {
            "unparsed_address": "Registry, Old College, City of Edinburgh, EDINBURGH, Midlothian. EH1 1AA UK",
            "country": "UK",
            "postcode": "EH1 1AA",
            "county": "Midlothian",
            "town": "EDINBURGH",
            "street": "Registry, Old College, City of Edinburgh",
        }

    def test_process_address_empty(self):
        # Given
        addresses = []

        # When
        result = contact._process_address(addresses)

        # Then
        assert isinstance(result, dict)
        assert all(key.__class__ is str for key in result.keys())
        assert all(value is None for value in result.values())
        assert [*result.keys()] == ["unparsed_address", "country", "postcode", "county", "town", "street"]
        assert result == {
            "unparsed_address": None,
            "country": None,
            "postcode": None,
            "county": None,
            "town": None,
            "street": None,
        }

    def test_process_phone_numbers(self):
        # Given
        numbers = [
            {"Number": "0", "IsMain": False},
            {"Number": "1", "IsMain": False},
            {"Number": "2", "IsMain": False},
            {"Number": "3", "IsMain": False},
            {"Number": "4", "IsMain": False},
            {"Number": "5", "IsMain": True},
        ]

        # When
        result = contact._process_phone_numbers(numbers)

        # Then
        assert result == "5"

    def test_process_phone_numbers_no_main(self):
        # Given
        numbers = [
            {"Number": "0", "IsMain": False},
            {"Number": "1", "IsMain": False},
            {"Number": "2", "IsMain": False},
            {"Number": "3", "IsMain": False},
            {"Number": "4", "IsMain": False},
            {"Number": "5", "IsMain": False},
        ]

        # When
        result = contact._process_phone_numbers(numbers)

        # Then
        assert result == "0"

    def test_process_phone_numbers_empty(self):
        # Given
        numbers = []

        # When
        result = contact._process_phone_numbers(numbers)

        # Then
        assert result == ""

    def test_process_email(self):
        # Given
        emails = [
            {"EmailAddress": "egg@GreenMidget.cafe", "IsMain": False},
            {"EmailAddress": "bacon@GreenMidget.cafe", "IsMain": False},
            {"EmailAddress": "sausage@GreenMidget.cafe", "IsMain": False},
            {"EmailAddress": "spam@GreenMidget.cafe", "IsMain": False},
            {"EmailAddress": "spamspam@GreenMidget.cafe", "IsMain": False},
            {"EmailAddress": "lobsterthermidor@GreenMidget.cafe", "IsMain": True},
        ]

        # When
        result = contact._process_email(emails)

        # Then
        assert result == "lobsterthermidor@GreenMidget.cafe"

    def test_process_email_no_main(self):
        # Given
        emails = [
            {"EmailAddress": "egg@GreenMidget.cafe", "IsMain": False},
            {"EmailAddress": "bacon@GreenMidget.cafe", "IsMain": False},
            {"EmailAddress": "sausage@GreenMidget.cafe", "IsMain": False},
            {"EmailAddress": "spam@GreenMidget.cafe", "IsMain": False},
            {"EmailAddress": "spamspam@GreenMidget.cafe", "IsMain": False},
            {"EmailAddress": "lobsterthermidor@GreenMidget.cafe", "IsMain": False},
        ]

        # When
        result = contact._process_email(emails)

        # Then
        assert result == "egg@GreenMidget.cafe"

    def test_process_email_empty(self):
        # Given
        emails = []

        # When
        result = contact._process_email(emails)

        # Then
        assert result == ""

    def test_process_misc_sections(self):
        # Given
        sections = [
            "A - 1",
            "B - 2",
            "C - 3",
            "D",
        ]

        # When
        result = contact._process_misc_sections(sections)

        # Then
        assert isinstance(result, dict)
        assert all(key.__class__ is str for key in result.keys())
        assert all(value.__class__ is str for value in result.values())
        assert [*result.keys()] == ["A", "B", "C", "D"]
        assert [*result.values()] == ["1", "2", "3", ""]
        assert result == {"A": "1", "B": "2", "C": "3", "D": ""}

    def test_process_misc_sections_empty(self):
        # Given
        sections = []

        # When
        result = contact._process_misc_sections(sections)

        # Then
        assert isinstance(result, dict)
        assert [*result.keys()] == []
        assert [*result.values()] == []
        assert result == {}

    def test_parse_iso_date(self):
        # Given
        date = "2000-01-01"

        # When
        result = contact._parse_iso_date(date)

        # Then
        assert result == datetime.date(2000, 1, 1)

    def test_parse_iso_date_none(self):
        # Given
        date = None

        # When
        result = contact._parse_iso_date(date)

        # Then
        assert result is None



import pytest
import uvicorn

from tests.util.fake_compass import app
from tests.util.http import Server


@pytest.fixture(scope="session")
def server():
    print("Starting uvicorn server!")
    config = uvicorn.Config(app, host="127.0.0.1", port=4200, log_level="info")
    server = Server(config=config)
    with server.run_in_thread():
        yield



from __future__ import annotations

from typing import Literal, Optional, TypedDict

# sync with _scrapers/member_profile.py::ROLE_STATUSES
TYPES_ROLE_STATUS = Literal["Cancelled", "Closed", "Full", "Pre provisional", "Provisional"]
TYPES_SEX = Literal["Male", "Female", "Unknown"]
TYPES_ETHNICITY = Literal[
    "1.English/Welsh/Scottish/Northern Irish/British",
    "2.Irish",
    "3.Gypsy or Irish Traveller",
    "4.Any other White background",
    "5.White and Black Caribbean",
    "6.White and Black African",
    "7.White and Asian",
    "8.Any other mixed or Multiple ethic group",
    "9.Indian",
    "10.Pakistani",
    "11.Bangladeshi",
    "12.Chinese",
    "13.Any other Asian Background",
    "14.African",
    "15.Caribbean",
    "16.Any other Black/African/Caribbean background",
    "17.Arab",
    "18.Other",
    "19.Prefer not to say",
]
TYPES_RELIGION = Literal[
    "Buddhist",
    "Christian (including all Christian denominations)",
    "Hindu",
    "Jewish",
    "Muslim",
    "Sikh",
    "Any other religion (please specify)",
    "No religion",
    "Prefer not to say",
]
TYPES_OCCUPATION = Literal[
    "Employed",
    "Unemployed",
    "Retired (whether receiving a pension or not)",
    "Student",
    "Long term sick or disabled",
    "Looking after home of family",
    "Other",
]
TYPES_ROLE_CLASS = Literal[
    "Administrator",
    "Advisor",
    "Assessor",
    "Co-ordinator",
    "Commissioner",
    "Committee",
    "Helper",
    "Honorary",
    "Leader",
    "Manager",
    "Secretary",
    "Staff",
    "System Role",
    "Supporter",
    "Trainer",
    "Default role class",
]
TYPES_REFERENCES = Literal[
    "Not Complete",
    "Not Required",
    "References Requested",
    "References Satisfactory",
    "References Unsatisfactory",
]
TYPES_DISCLOSURES_APPOINTMENT = Literal[
    "Application submitted - in progress",
    "Disclosure Issued",
    "ID check required",
    "No Disclosure",
    "No disclosure",
    "Expired",
    "Application Withdrawn",
    "Final applicant information required",
    "Application received at Atlantic Data",
    "Issued : ",
    "ID selection required",
    "Form returned to District",
    "Application form required",
    "Form returned to applicant",
    "Application received at HQ",
]  # Disclosure statuses in role details popup
TYPES_LEARNING_METHOD = Literal[
    "Course",
    "DVD/Video",
    "E-Learning",
    "External Course",
    "Independent Study",
    "On the Job",
    "One to One",
    "Practical",
    "Small Group",
    "Workbook",
    "Your prior learning/experience recognised",
    "Other",
    # First Aid
    "Life Support",
    "Major illness",
    "Trauma and injury",
    # Manager and Supporter: Induction (rare)
    "RST Induction [England only]",
    # Manager and Supporter: Skills Courses
    "Skills Course - Achieving Growth",
    "Skills Course - Meeting the Challenges",
    "Skills Course - Skills of Management",
    # Manager and Supporter: Independent Learning
    "IL - Building Effective Teams",
    "IL - Dealing with Difficult Situations",
    "IL - Decision Making",
    "IL - Enabling Change",
    "IL - Financial and Physical Resources",
    "IL - Finding, Appointing and Welcoming Volunteers",
    "IL - Getting the Word Out",
    "IL - Keeping, Developing and Managing Volunteers",
    "IL - Leading Local Scouting",
    "IL - Managing Time and Personal Skills",
    "IL - Planning For Growth",
    "IL - Project Management",
    "IL - Safety for Managers and Supporters",
    "IL - Supporting the Adult Training Scheme",
]
TYPES_PERMIT = Literal[
    "Archery",
    "Bell Boating",
    "Canoeing",
    "Caving",
    "Climbing (Multi Pitch)",
    "Climbing and Abseiling",
    "Dinghy Sailing",
    "Dragon Boating",
    "Hill Walking",
    "Hovercrafting",
    "Ice Climbing",
    "Kayaking",
    "Keelboating",
    "Kite Surfing",
    "Mine Exploration",
    "Motor Cruising",
    "Mountain Biking",
    "Narrow Boating",
    "Nights Away",
    "Personal Watercraft (Jet Ski)",
    "Power Boating",
    "Pulling",
    "Rafting (Traditional)",
    "Rowing and Sculling",
    "Scuba Diving",
    "Skiing",
    "Snorkelling",
    "Snowboarding",
    "Snowsports",
    "Water Skiing",
    "White Water Rafting",
    "Windsurfing",
    "Yachting",
]
TYPES_PERMIT_CATEGORIES = Literal[
    "B1 Waters",
    "B2 Waters",
    "B2+ Waters",
    "B3 Waters",
    "C Waters",
    "Campsite",
    "Greenfield",
    "Indoor",
    "Lightweight Expedition",
    "Open Inland B1 Waters",
    "Open Inland B2 Waters",
    "Open Inland B3 Waters",
    "River B1 Waters",
    "Sea B1 Waters",
    "Sea B2 Waters",
    "Surf B2 Waters",
    "With Compound Bows",
    "Without Compound Bows",
]
TYPES_PERMIT_TYPE = Literal["Leadership", "Supervisory", "Personal"]
# https://heritage.scouts.org.uk/wp-content/uploads/2019/07/Awards-Decorations-and-their-Emblems-of-The-Scout-Association-formerly-the-Boy-Scout-Association.pdf
TYPES_AWARD_TYPE = Literal[
    "Chief Scout's 5 years Service Award",
    "Chief Scout's 10 years Service Award",
    "Chief Scout's 15 years Service Award",
    "Chief Scout's 20 years Service Award",
    "Chief Scout's 25 years Service Award",
    "Chief Scout's 30 years Service Award",
    "Chief Scout's 40 years Service Award",
    "Chief Scout's 50 years Service Award",
    "Chief Scout's 60 years Service Award",
    "Chief Scout's 70 years Service Award",
    # Local / less formal awards
    "Commissioners Commendation",
    # Formal awards process - lower good service
    "Chief Scout's Commendation for Good Service",
    "Award for Merit",
    "Bar to the Award for Merit",
    "Silver Acorn",
    # Formal awards process - higher good service
    "Bar to the Silver Acorn",
    "Silver Wolf",
    # Meritorious Conduct Awards
    "Chief Scout's Commendation for Meritorious Conduct",
    "The Medal for Meritorious Conduct",
    # Gallantry Awards
    "The Gilt Cross",
    "The Silver Cross",
    "The Bronze Cross",
    # Other Awards
    "The Cornwell Scout Badge",
    "The Chief Scout's Personal Award",
    # Non-TSA Awards
    "Bronze Wolf",
    # Past Awards (not in the current awards scheme)
    "Medal of Merit",  # stopped being awarded in 2007
    "Bar to the Medal of Merit",  # stopped being awarded in 2007
    "Chief Scout's Commendation for Gallantry",  # stopped being awarded in 2006-04
    "Certificate for Meritorious Conduct",  # stopped being awarded in 2006-04
    "Bar to the Medal for Meritorious Conduct",
]
TYPES_DISCLOSURES_COUNTRIES = Optional[
    Literal[
        "England & Wales",
        "Northern Ireland",
        "Overseas",
        "Scotland",
        "The Scout Association",
        "Test Country",
        "UK",
    ]
]
TYPES_DISCLOSURE_PROVIDERS = Literal[
    "Access NI",
    "Atlantic Data",
    "DBS Paper Application",
    "Disclosure Scotland",
    "Local Check",  # BSO
    "Other",
    "UKHQ",
]
TYPES_DISCLOSURE_STATUSES = Literal[
    "Application Withdrawn",
    "Application received at HQ",
    "Application submitted - in progress",
    "Disclosure Expired",
    "Disclosure Issued",
    "Expired",
    "Final applicant information required",  # BSO
    "ID check required",
    "ID selection required",
    "No Disclosure",
    "Application received at Atlantic Data",
    "Application form required",
    "Form returned to District",
    "Issued",
    "Form returned to applicant",
]  # Disclosure statuses in disclosures tab


class AddressData(TypedDict):
    unparsed_address: Optional[str]
    country: Optional[str]
    postcode: Optional[str]
    county: Optional[str]
    town: Optional[str]
    street: Optional[str]



import httpx
import pytest

import compass.core as ci
from compass.core._scrapers import reports
from compass.core.settings import Settings

base_url = "http://127.0.0.1:4200"


class TestReports:
    def test_report_number(self):
        # Given
        report_type = "Awards Report"
        hierarchy_level = "District"

        # When
        out = reports._report_number(report_type, hierarchy_level)

        # Then
        assert out == 94

    def test_report_number_invalid_type(self):
        # Given
        report_type = "Youth Members Report"
        hierarchy_level = "District"

        # Then
        with pytest.raises(ci.CompassReportError, match="Youth Members Report is not a valid report type"):
            # When
            reports._report_number(report_type, hierarchy_level)

    def test_report_number_invalid_level(self):
        # Given
        report_type = "Awards Report"
        hierarchy_level = "Group"

        # Then
        with pytest.raises(ci.CompassReportError, match="Requested report does not exist for hierarchy level: Group."):
            # When
            reports._report_number(report_type, hierarchy_level)

    def test_extract_form_data(self):
        # TODO -- lxml
        pass

    def test_form_data_appointments(self):
        # TODO -- lxml
        pass

    def test_parse_drop_down_list(self):
        # TODO -- lxml
        pass

    def test_extract_report_export_url_base(self):
        # Given
        Settings.base_url = base_url
        haystack = """Lorem ipsum dolor sit amet, "ExportUrlBase":"run-report-export" consectetur adipiscing elit. Phasellus """

        # When
        out = reports._extract_report_export_url_base(haystack)

        # Then
        assert out == "http://127.0.0.1:4200/run-report-export"

        # TODO test unhappy path, unicode escape stuff

    def test_error_status(self):
        # Given
        transport = httpx.MockTransport(lambda _: httpx.Response(500))
        client = httpx.Client(transport=transport)

        # When
        response = client.get("http://a-uri-that-returns-an-error-status")

        # Then
        with pytest.raises(ci.CompassNetworkError, match="Request to Compass failed!"):
            reports._error_status(response)



import datetime

from compass.core._scrapers import role_detail


class TestRoleDetail:
    def test_extract_line_manager(self):
        # TODO -- lxml
        pass

    def test_extract_disclosure_date(self):
        # Given
        haystack = "Disclosure Issued : 01 Jan 2000"

        # When
        result = role_detail._extract_disclosure_date(haystack)

        # Then
        assert result == ("Disclosure Issued", datetime.date(2000, 1, 1))

    def test_extract_disclosure_date_invalid_prefix(self):
        # Given
        haystack = "The Killer Joke : 8 July 1944"

        # When
        result = role_detail._extract_disclosure_date(haystack)

        # Then
        assert result == (haystack, None)

    def test_extract_disclosure_date_invalid_suffix(self):
        # Given
        haystack = "Disclosure Issued : "

        # When
        result = role_detail._extract_disclosure_date(haystack)

        # Then
        assert result == ("Disclosure Issued", None)

    def test_extract_disclosure_date_empty(self):
        # Given
        haystack = ""

        # When
        result = role_detail._extract_disclosure_date(haystack)

        # Then
        assert result == (None, None)

    def test_process_hierarchy(self):
        # TODO -- lxml
        pass

    def test_process_getting_started(self):
        # TODO -- lxml
        pass



from typing import Literal, Optional, TypedDict, Union

# TYPES_ENDPOINT_LEVELS values are meaningful values as they become the API endpoint paths
TYPES_ENDPOINT_LEVELS = Literal[
    "countries",
    "hq_sections",
    "regions",
    "country_sections",
    "counties",
    "region_sections",
    "districts",
    "county_sections",
    "groups",
    "district_sections",
    "group_sections",
]
TYPES_SECTION = Literal[
    "EY Pilot",
    "Squirrel Scout",
    "Beavers",
    "Cubs",
    "Scouts",
    "Explorers",
    "Network",
    "ASU",
    "Other",
]
TYPES_UNIT_LEVELS = Literal["Group", "District", "County", "Region", "Country", "Organisation"]
TYPES_SECTION_LEVELS = Literal[
    "Group Section",
    "District Section",
    "County Section",
    "Regional Section",
    "Country Section",
    "Organisation Section",
]
TYPES_HIERARCHY_LEVELS = Union[TYPES_UNIT_LEVELS, TYPES_SECTION_LEVELS]
_TYPES_LEVEL = tuple[Optional[TYPES_UNIT_LEVELS], Optional[TYPES_ENDPOINT_LEVELS], TYPES_ENDPOINT_LEVELS]


class _HierarchyState(TypedDict, total=False):
    unit_id: int
    name: Optional[str]
    # organisation: int  # Always 10000001; mildly redundant
    country: int
    region: int
    county: int
    district: int
    group: int



import datetime

import pytest

from compass.core.util import type_coercion


class TestTypeCoercion:
    def test_maybe_int_int(self):
        # Given
        data = 123

        # When
        result = type_coercion.maybe_int(data)
        result_str = type_coercion.maybe_int(str(data))

        # Then
        assert result == data
        assert result_str == data

    def test_maybe_int_str(self):
        # Given
        data = "abc"

        # When
        result = type_coercion.maybe_int(data)

        # Then
        assert result is None

    def test_parse_month_short(self):
        # Given
        data = "01 Jan 2000"

        # When
        result = type_coercion.parse_date(data)

        # Then
        assert isinstance(result, datetime.date)
        assert result == datetime.date(2000, 1, 1)

    def test_parse_month_long(self):
        # Given
        data = "01 January 2000"

        # When
        result = type_coercion.parse_date(data)

        # Then
        assert isinstance(result, datetime.date)
        assert result == datetime.date(2000, 1, 1)

    def test_parse_non_date(self):
        # Given
        data = "abc"

        # Then
        with pytest.raises(ValueError, match=f"Parsing string `{data}` into a date failed!"):
            # When
            type_coercion.parse_date(data)

    def test_parse_empty(self):
        # Given
        data = ""

        # When
        result = type_coercion.parse_date(data)

        # Then
        assert result is None



from collections.abc import Iterator
import inspect

from compass.core import hierarchy
import compass.core as ci


class TestHierarchy:
    def test_flatten_hierarchy(self):
        # Given
        hierarchy_data = ci.UnitData(
            unit_id=10001234,
            name=None,
            level="District",
            child=[
                ci.DescendantData(
                    unit_id=10050000,
                    name="1st Whitehall",
                    level="Group",
                    child=None,
                    sections=[
                        ci.HierarchySection(unit_id=10050100, name="1st Whitehall Beavers", section_type="Beavers"),
                        ci.HierarchySection(unit_id=10050200, name="1st Whitehall Cubs", section_type="Cubs"),
                        ci.HierarchySection(unit_id=10050300, name="1st Whitehall Scouts", section_type="Scouts"),
                    ],
                ),
                ci.DescendantData(
                    unit_id=10060000,
                    name="2nd Pimlico",
                    level="Group",
                    child=None,
                    sections=[
                        ci.HierarchySection(unit_id=10060100, name="Pimlico Beaver Scouts", section_type="Beavers"),
                        ci.HierarchySection(unit_id=10060200, name="Pimlico Cubs", section_type="Cubs"),
                        ci.HierarchySection(unit_id=10060300, name="Pimlico Cub Scouts", section_type="Cubs"),
                        ci.HierarchySection(unit_id=10060400, name="Pimlico Scouts", section_type="Scouts"),
                    ],
                ),
            ],
            sections=[
                ci.HierarchySection(unit_id=10007100, name="M25 Scout Network", section_type="Network"),
                ci.HierarchySection(unit_id=10007200, name="District Active Support Unit", section_type="ASU"),
                ci.HierarchySection(unit_id=10007300, name="Londinium Explorers", section_type="Explorers"),
                ci.HierarchySection(unit_id=10007400, name="YL ESU", section_type="Explorers"),
            ],
        )

        # When
        result = hierarchy._flatten_hierarchy(hierarchy_data)

        # Then
        assert inspect.isgenerator(result)
        assert isinstance(result, Iterator)
        result = [*result]
        assert result == [
            {"unit_id": 10001234, "name": None, "section": False, "district": 10001234},
            {"unit_id": 10050000, "name": "1st Whitehall", "section": False, "district": 10001234, "group": 10050000},
            {"unit_id": 10050100, "name": "1st Whitehall Beavers", "section": True, "district": 10001234, "group": 10050000},
            {"unit_id": 10050200, "name": "1st Whitehall Cubs", "section": True, "district": 10001234, "group": 10050000},
            {"unit_id": 10050300, "name": "1st Whitehall Scouts", "section": True, "district": 10001234, "group": 10050000},
            {"unit_id": 10060000, "name": "2nd Pimlico", "section": False, "district": 10001234, "group": 10060000},
            {"unit_id": 10060100, "name": "Pimlico Beaver Scouts", "section": True, "district": 10001234, "group": 10060000},
            {"unit_id": 10060200, "name": "Pimlico Cubs", "section": True, "district": 10001234, "group": 10060000},
            {"unit_id": 10060300, "name": "Pimlico Cub Scouts", "section": True, "district": 10001234, "group": 10060000},
            {"unit_id": 10060400, "name": "Pimlico Scouts", "section": True, "district": 10001234, "group": 10060000},
            {"unit_id": 10007100, "name": "M25 Scout Network", "section": True, "district": 10001234},
            {"unit_id": 10007200, "name": "District Active Support Unit", "section": True, "district": 10001234},
            {"unit_id": 10007300, "name": "Londinium Explorers", "section": True, "district": 10001234},
            {"unit_id": 10007400, "name": "YL ESU", "section": True, "district": 10001234},
        ]



from pathlib import Path

Path(__file__, "..", "version.txt").resolve().write_text(
    # read version file
    Path(__file__, "..", "..", "compass", "core", "__version__.py").resolve().read_text(encoding="utf-8")
    # get first line
    .split("\n")[0]
    # remove lead and tail
    .removeprefix('__version__ = "').removesuffix('"'),
    encoding="utf-8",
)



import os
import uuid
import warnings

from starlette.applications import Starlette
from starlette.requests import Request
from starlette.responses import HTMLResponse

from compass.core.logon import Settings

# ASP.NET id is 24 characters, lower-case alpha and digits
asp_net_id = os.urandom(12).hex()
id_map: dict[str, str] = {}  # ASP.NET id -> session id
username_map: dict[str, str] = {}  # ASP.NET id -> username
session_id_default = "d6c76537-1b6c-3910-c3d4-d21d4e6453a6"

app = Starlette()


def generate_asp_net_id() -> str:
    return os.urandom(12).hex()


@app.route("/", methods=["head"])
async def get_root(_request: Request) -> HTMLResponse:
    _asp_net_id = generate_asp_net_id()
    id_map[_asp_net_id] = str(uuid.uuid4())
    response = HTMLResponse(b"")
    response.set_cookie("ASP.NET_SessionId", _asp_net_id)
    return response


@app.route("/_testing/no-cookie", methods=["head"])
async def get_testing_no_cookie(_request: Request) -> HTMLResponse:
    return HTMLResponse(b"")


@app.route("/Login.ashx", methods=["post"])
async def post_login_ashx(request: Request) -> HTMLResponse:
    if request.headers.get("Referer") != f"{Settings.base_url}/login/User/Login":
        # If the Referer header is not set properly, Compass silently fails and returns the login page
        return HTMLResponse(
            b'<!DOCTYPE html>\r\n<html lang="en">\r\n<head>\r\n    <meta charset="utf-8" />\r\n    <meta name="viewport" content="width=device-width, initial-scale=1, maximum-scale=1">\r\n    <title>The Scout Association</title>\r\n    <link rel="shortcut icon" media="all" type="image/x-icon" href="/login/Images/ico_compass.ico" />\r\n    <link rel="icon" media="all" type="image/vnd.microsoft.icon" href="/login/Images/ico_compass.ico" />\r\n    <!--[if IE]><script src="~/Scripts/html5.js"></script><![endif]-->\r\n\r\n    <script type="text/javascript">\r\n        //Break out of iframe\r\n        if (top.location !== location) {\r\n            top.location.href = document.location.href;\r\n        }\r\n    </script>\r\n    <link href="/login/Content/css?v=VZKccHPLGSKNFic6yG06sWCvfqCdbhlOngzX1wgfnHc1" rel="stylesheet"/>\r\n\r\n    <script src="/login/bundles/modernizr?v=qVODBytEBVVePTNtSFXgRX0NCEjh9U_Oj8ePaSiRcGg1"></script>\r\n\r\n</head>\r\n'
            b'<body>\r\n    <div id="background">\r\n        <div id="wrapper">\r\n            <div id="header">\r\n                <div id="logo">\r\n                    <img src="/login/Images/Template/The_Scout_Association.png" alt="The Scout Association Logo" />\r\n                </div>\r\n                <div id="maintitle">\r\n                    <h1>Login</h1>\r\n                </div>\r\n            </div>\r\n            <div id="content">\r\n                <div class="login-form-intro-text">\r\n    <p>\r\n        <i class="fa fa-info-circle scouts-colour"></i> Check out the <a href="http://compasssupport.scouts.org.uk/?page_id=59542" target="_blank" title="Warning, opens in a new window!">Support Site</a>\r\n        including video tutorials, user guides, <a href="http://compasssupport.scouts.org.uk/?page_id=59643" target="_blank" title="Warning, opens in a new window!">FAQs</a> and our \xe2\x80\x98'
            b'<a href="http://compasssupport.scouts.org.uk/?page_id=59537" target="_blank" title="Warning, opens in a new window!">Latest updates</a>\xe2\x80\x99 blog.\r\n    </p>\r\n    <p><i class="fa fa-exclamation-circle scouts-colour"></i> <a href="Register">Register</a> for Compass before you log in for the first time.\r\n    </p>\r\n</div>\r\n\r\n<!--Note, the compassMessages div is meant to be used to display external messages. The idea was that the messages would be hosted in a different server\r\n(members.scouts.org.uk) so that we could control the messages without having to deploy this page in Compass. Not working as of 17/02/2016 because\r\n we did not actaully complete it and approve it for go live.-->\r\n<div id="compassMessages">\r\n    <h2 id="compassMessagesTitle"></h2>\r\n    <span id="compassMessagesContent"></span>\r\n</div>\r\n\r\n\r\n\r\n'
            b'<form class="form-horizontal" role="form" action="https://compass.scouts.org.uk/Login.ashx" autocomplete="off" method="post" id="logonForm" data-parsley-validate>\r\n    <div class="form-group">\r\n        <label class="control-label col-sm-2" for="EM">Username</label>\r\n        <div class="col-sm-4">\r\n            <input type="text" class="form-control" id="EM" name="EM" placeholder="Enter username"\r\n                   required\r\n                   maxlength="80"\r\n                   data-parsley-required-message="Username is required"\r\n                   data-parsley-maxlength="80"\r\n                   data-parsley-maxlength-message="Username must not exceed 80 characters in length." />\r\n        </div>\r\n    </div>\r\n\r\n    <div class="form-group">\r\n        <label class="control-label col-sm-2" for="PW">Password</label>\r\n        <div class="col-sm-4">\r\n            '
            b'<input type="password" class="form-control" id="PW" name="PW" placeholder="Enter password"\r\n                   required\r\n                   maxlength="100"\r\n                   data-parsley-required-message="Password is required"\r\n                   data-parsley-maxlength="100"\r\n                   data-parsley-maxlength-message="Password must not exceed 100 characters in length.">\r\n        </div>\r\n    </div>\r\n\r\n    <div class="form-group">\r\n        <div class="col-sm-offset-2 col-sm-10">\r\n            <input type="hidden" name="ON" value="10000001" />\r\n            <button type="submit" class="btn btn-primary">Log in</button>\r\n        </div>\r\n    </div>\r\n</form>\r\n\r\n\r\n<div class="manageAccountLinksContainer">\r\n    <ul>\r\n        <li><a href="/login/User/Register">Register for your account</a> '
            b'(<a href="http://compasssupport.scouts.org.uk/?guide-cat=using-compass" target="_blank" title="Warning, opens in a new window!">Need help?</a>)</li>\r\n        <li> <a href="/login/User/ForgotUsername">Retrieve forgotten username</a> (<a href="http://compasssupport.scouts.org.uk/?guide-cat=using-compass" target="_blank" title="Warning, opens in a new window!">Need help?</a>)</li>\r\n        <li> <a href="/login/User/ForgotPassword">Reset your password</a> (<a href="http://compasssupport.scouts.org.uk/?guide-cat=using-compass" target="_blank" title="Warning, opens in a new window!">Need help?</a>)</li>\r\n    </ul>\r\n</div>\r\n\r\n\r\n\r\n            </div>\r\n            <div id="footer">\r\n                <!--<p>This is the Footer</p>-->\r\n            </div>\r\n        </div>\r\n    </div>\r\n    '
            b'<script type="text/javascript">\r\n\r\n        if (typeof console === "undefined") {\r\n            console = { log: function () { } };\r\n        }\r\n\r\n        window.ScoutLoginApplication = {};\r\n        (function (tsaApp) {\r\n            var rootPath = \'\';\r\n            tsaApp.rootPath = rootPath;\r\n        }(window.ScoutLoginApplication));\r\n        ScoutLoginApplication.rootPath = \'/login/\';\r\n        var applicationMessagesArray = new Array();\r\n            \r\n        applicationMessagesArray["RegistrationGenericErrorMsg"] = "Sorry, there was a problem registering you. You did not provide the correct information in one or more of the fields below. Please check your answers carefully. If this problem persists please call the Scout Information Centre on 0845 300 1818 or 020 8433 7100 for help."\r\n        \r\n            \r\n        applicationMessagesArray["EnableDatePickerWidgetOnAllForms"] = "True"'
            b'\r\n        \r\n            \r\n        applicationMessagesArray["InvalidUKPostCodePatternErrMsg"] = "Postcode does not match the required pattern. Take care to ensure that all letters are in UPPER CASE and that there is a space between the outer and inner parts of the post code e.g. E4 7QW instead of e47qw."\r\n        \r\n            \r\n        applicationMessagesArray["ValidUKPostCodePatternMsg"] = "Postcode pattern is valid."\r\n        \r\n            \r\n        applicationMessagesArray["ValidPasswordPatternMsg"] = "Password pattern is valid."\r\n        \r\n            \r\n        applicationMessagesArray["InvalidPasswordPatternErrMsg"] = "Password does not match the required pattern."\r\n        \r\n            \r\n        applicationMessagesArray["InvalidUsernamePatternErrMsg"] = "Username does not match allowed pattern"\r\n        \r\n            \r\n        '
            b'applicationMessagesArray["ValidUsernamePatternMsg"] = "Username pattern is valid"\r\n        \r\n\r\n        ScoutLoginApplication.applicationMessages = applicationMessagesArray;\r\n\r\n    </script>\r\n    \r\n    <script src="/login/bundles/logonjs?v=GIPO7plg5Q-bbDF8BW0hg7usbTfLYkbl72xdM3YPvts1"></script>\r\n\r\n\r\n</body>\r\n</html>\r\n'
        )
    valid_users = {
        b"EM=username&PW=password&ON=10000001": "username",
        b"EM=no_role_number&PW=password&ON=10000001": "no_role_number",
    }
    body = await request.body()
    if body not in valid_users:
        # If credentials are not correct, redirect to Failed Login page
        return HTMLResponse(
            b"<head><title>Compass - Failed Login</title><link rel='shortcut icon' type='image/vnd.microsoft.icon' href='https://compass.scouts.org.uk/Images/core/ico_compass.ico' sizes='16x16 24x24 32x32 48x48'></head><body onload='window.location.href=\"\"'></body>"
        )

    # store the requestor's id
    _asp_net_id = request.cookies["ASP.NET_SessionId"]
    username_map[_asp_net_id] = valid_users[body]

    # If all OK, return system startup page
    return HTMLResponse(
        b"<head><title>Compass - System Startup</title><link rel='shortcut icon' type='image/vnd.microsoft.icon' href='https://compass.scouts.org.uk/Images/core/ico_compass.ico' sizes='16x16 24x24 32x32 48x48'></head><body onload='window.location.href=\"https://compass.scouts.org.uk/ScoutsPortal.aspx\"'></body>"
    )


@app.route("/MemberProfile.aspx", methods=["get"])
async def get_member_profile(request: Request) -> HTMLResponse:
    if "TAB" not in request.query_params:
        warnings.warn("Didn't pass TAB", RuntimeWarning)
    page = request.query_params.get("Page")
    if page is None:
        pass  # TODO return Personal Details tab
    elif page == "ROLES":
        _asp_net_id = request.cookies["ASP.NET_SessionId"]
        username = username_map[_asp_net_id]
        role_number = {
            "username": b"~Master.User.MRN#9000000",
            "no_role_number": b"",
        }
        return HTMLResponse(
            b"<!DOCTYPE html><html xmlns='http://www.w3.org/1999/xhtml' id='mainhtml'><head><title>Scout - 10000000 John Smith</title></head><body><form name='aspnetForm' method='post' action='./MemberProfile.aspx?Page=ROLES&amp;TAB' id='aspnetForm' autocomplete='off'><div id='TSA'><input name='ctl00$_POST_CTRL' type='hidden' id='ctl00__POST_CTRL' value='Nav.Action#None~Nav.StartNo#-1~Nav.StartPage#3~Page.UseCN#10000000~CRUD.MDIS#R~CRUD.ROLES#R~CRUD.PEMD#R~CRUD.MMMD#R~CRUD.MVID#R~CRUD.PERM#R~CRUD.TRN#CRD~Page.HideBadges#Y~Page.Croc#OK~Page.HideNominations#Y~User.IsMe#Y~Page.CanDeleteOGLHrs#N~Page.FoldName#MP_~Master.SSO.ON#10000001~Master.User.CN#10000000"
            + role_number[username]
            + b"~Master.User.ON#10000001~Master.User.LVL#ORG~Master.User.JK#9b65d68f4aca0138b5bae4492e7cdfae220a834e3e69731d996be1ddbb496d32fd29497d4f9729525c9fbc77666bb520ea214c0802ea22b958e6ae525224fd15~Master.Const.Wales#10000007~Master.Const.Scotland#10000005~Master.Const.OverSeas#10000002~Master.Const.HQ#10000001~Master.Sys.SessionID#"
            + id_map[_asp_net_id].encode()
            + b"~Master.Sys.SafeJSON#Y~Master.Sys.WebPath#https%3a%2f%2fcompass.scouts.org.uk%2fJSon.svc%2f~Master.Sys.TextSize#1~Master.Sys.Timout#1729000~Master.Sys.REST#Y~Master.Sys.HardTime#20:18.28~Master.Sys.HardExpiry#6 hours~Master.Sys.TimeoutExtension#10~Master.Sys.Ping#300000~Const.Sys.STO#5~Const.Sys.STO_ASK#0'/></div><div id='ctl00_Popup_UpdatePanel'><div id='mstr_container'><div id='mstr_panel'><div id='mstr_scroll' style='position: relative; height: 50000px;'><div id='mstr_work'><div id='mpage3' class='mpage' style='visibility:hidden;'><div id='divProfile16'><table class='msTable' id='tbl_p3_roles'><tbody>"
            + b"<tr class='msTRRL msTR' data-pk='9000000'><td class='tdData'><a class='VIEWRL' data-ng_id='1' Title='Administrator' href='#'>Regional Administrator</a></td><td class='tdData'><label>Administrator</label></td><td class='tdData'><a class='VIEWHR' data-ng_id='10000069' data-lvl='REG' href='#'>Wessex</a></td><td class='tdData' style='white-space: nowrap;'><label>10 October 2019</label></td><td class='tdData'></td><td class='tdData' style='white-space: nowrap;'><label class='ll_role_Status'>Full</label><br/>Review Due<br/>30 Sep 2022</td><td class='tdData noPrint SortOpts' style='text-align:right '><label style='display:none;width:1px;' class='ORD' data-db='1'>1</label><input type='button' class='VIEWROLE VIEWMRN' value='View'/><div class='EDITROLE' style='width:260px;display:none;'><input type='button' class='RoleTop' value='Make Primary'/><input type='button' class='RoleUp' value='Up'/><input type='button' class='RoleDown' value='Down'/></div></td></tr>"
            + b"<tr class='msTRRL msTR' data-pk='6857721'><td class='tdData'><a class='VIEWRL' data-ng_id='24' Title='Scout Council Member' href='#'>TSA Council Member - Nominated Member (18-24)</a></td><td class='tdData'><label>Committee</label></td><td class='tdData'><a class='VIEWHR' data-ng_id='10000001' data-lvl='ORG' href='#'>The Scout Association</a></td><td class='tdData' style='white-space: nowrap;'><label>14 December 2018</label></td><td class='tdData'></td><td class='tdData' style='white-space: nowrap;'><label class='ll_role_Status'>Full</label></td><td class='tdData noPrint SortOpts' style='text-align:right '><label style='display:none;width:1px;' class='ORD' data-db='2'>2</label><input type='button' class='VIEWROLE VIEWMRN' value='View'/><div class='EDITROLE' style='width:260px;display:none;'><input type='button' class='RoleTop' value='Make Primary'/><input type='button' class='RoleUp' value='Up'/><input type='button' class='RoleDown' value='Down'/></div></td></tr>"
            + b"<tr class='msTRRL msTR ROLE_HIDEME' data-pk='3582685'><td class='tdData'><a class='VIEWRL' data-ng_id='74' Title='Scout Network Member' href='#'>County Scout Network Member</a></td><td class='tdData'><label>Supporter</label></td><td class='tdData'><a class='VIEWHR' data-ng_id='10000421' data-lvl='CTST' href='#'>Scout Network @ Trumptonshire</a></td><td class='tdData' style='white-space: nowrap;'><label>16 April 2018</label></td><td class='tdData' style='white-space: nowrap;'><label>18 April 2018</label></td><td class='tdData' style='white-space: nowrap;'><label class='ll_role_Status'>Closed</label></td><td class='tdData noPrint SortOpts' style='text-align:right '><label style='display:none;width:1px;' class='ORD' data-db='10003'>10003</label><input type='button' class='VIEWROLE VIEWMRN' value='View'/></td></tr>"
            + b"<tr class='msTRRL msTR ROLE_HIDEME' data-pk='4801342'><td class='tdData'><a class='VIEWRL' data-ng_id='22' Title='Executive Committee Member' href='#'>County Executive Committee Member</a></td><td class='tdData'><label>Committee</label></td><td class='tdData'><a class='VIEWHR' data-ng_id='10000420' data-lvl='CNTY' href='#'>Trumptonshire</a></td><td class='tdData' style='white-space: nowrap;'><label>25 September 2018</label></td><td class='tdData' style='white-space: nowrap;'><label>29 September 2020</label></td><td class='tdData' style='white-space: nowrap;'><label class='ll_role_Status'>Closed</label></td><td class='tdData noPrint SortOpts' style='text-align:right '><label style='display:none;width:1px;' class='ORD' data-db='10004'>10004</label><input type='button' class='VIEWROLE VIEWMRN' value='View'/></td></tr>"
            + b"<tr class='msTRRL msTR ROLE_HIDEME' data-pk='1158267'><td class='tdData'><a class='VIEWRL' data-ng_id='7' Title='Boat Inspector' href='#'>Regional Boat Inspector</a></td><td class='tdData'><label>Advisor</label></td><td class='tdData'><a class='VIEWHR' data-ng_id='10000069' data-lvl='REG' href='#'>Wessex</a></td><td class='tdData' style='white-space: nowrap;'><label>01 April 2020</label></td><td class='tdData'></td><td class='tdData' style='white-space: nowrap;'><label class='ll_role_Status'>Cancelled</label></td><td class='tdData noPrint SortOpts' style='text-align:right '><label style='display:none;width:1px;' class='ORD' data-db='10005'>10005</label><input type='button' class='VIEWROLE VIEWMRN' value='View'/></td></tr>"
            + b"</tbody></table></div></div></div></div></div></div></div></form></body></html>"
        )
    # TODO error conditions



import datetime

import pytest

import compass.core as ci
from compass.core._scrapers import member_profile


class TestMemberProfile:
    def test_process_address(self):
        # Given
        address = "Registry, Old College, City of Edinburgh, EDINBURGH, Midlothian. EH1 1AA UK"

        # When
        result = member_profile._process_address(address)

        # Then
        assert isinstance(result, dict)
        assert all(key.__class__ is str for key in result.keys())
        assert all(value.__class__ is str for value in result.values())
        assert [*result.keys()] == ["unparsed_address", "country", "postcode", "county", "town", "street"]
        assert result == {
            "unparsed_address": "Registry, Old College, City of Edinburgh, EDINBURGH, Midlothian. EH1 1AA UK",
            "country": "UK",
            "postcode": "EH1 1AA",
            "county": "Midlothian",
            "town": "EDINBURGH",
            "street": "Registry, Old College, City of Edinburgh",
        }

    def test_process_extra(self):
        # Given
        extra = "A - 1"

        # When
        result = member_profile._process_extra(extra)

        # Then
        assert result == ("A", "1")

    def test_process_extra_blank_detail(self):
        # Given
        extra = "D"

        # When
        result = member_profile._process_extra(extra)

        # Then
        assert result == ("D", None)

    def test_process_misc_sections(self):
        # TODO - lxml
        pass

    def test_extract_primary_role(self):
        # Given
        primary_role = None
        role_title = "Section Leader [Primary]"

        # When
        result_title, result_role = member_profile._extract_primary_role(role_title, primary_role)

        # Then
        assert result_role is True
        assert result_title == "Section Leader"

    def test_extract_primary_role_already_found(self):
        # Given
        primary_role = 99999999
        role_title = "some unchanged words"

        # When
        result_title, result_role = member_profile._extract_primary_role(role_title, primary_role)

        # Then
        assert result_role == 99999999
        assert result_title == "some unchanged words"

    def test_extract_review_date(self):
        # Given
        review = "Full Review Due 01 Jan 2000"

        # When
        result_role_status, result_review_date = member_profile._extract_review_date(review)

        # Then
        assert result_role_status == "Full"
        assert result_review_date == datetime.date(2000, 1, 1)

    def test_extract_review_date_ending(self):
        # Given
        review = "Full Ending 01 Jan 2000"

        # When
        result_role_status, result_review_date = member_profile._extract_review_date(review)

        # Then
        assert result_role_status == "Full"
        assert result_review_date == datetime.date(2000, 1, 1)

    @pytest.mark.parametrize("status", ["Cancelled", "Closed", "Full", "Pre provisional", "Provisional"])
    def test_extract_review_date_statuses(self, status):
        # Given
        review = status

        # When
        result_role_status, result_review_date = member_profile._extract_review_date(review)

        # Then
        assert result_role_status == status
        assert result_review_date is None

    def test_extract_review_date_invalid(self):
        # Given
        review = "invalid status"

        # Then
        with pytest.raises(ci.CompassError, match="Invalid value for review status 'invalid status'!"):
            # When
            member_profile._extract_review_date(review)

    def test_membership_duration(self):
        # Given
        data = [
            (datetime.date(2016, 4, 5), datetime.date(2020, 2, 7)),
            (datetime.date(2015, 1, 1), datetime.date(2021, 1, 1)),
            (datetime.date(2017, 1, 1), datetime.date(2021, 6, 1)),
            (datetime.date(2014, 7, 22), datetime.date(2015, 6, 10)),
            (datetime.date(2015, 1, 1), datetime.date(2015, 1, 1)),
            (datetime.date(2004, 5, 13), datetime.date(2007, 11, 1)),
            (datetime.date(1997, 7, 2), datetime.date(2007, 11, 1)),
            (datetime.date(1993, 2, 18), datetime.date(1997, 7, 2)),
            (datetime.date(1983, 11, 17), datetime.date(1985, 3, 23)),
        ]

        # When
        result = member_profile._membership_duration(data)

        # Then
        assert result == 22.916

    def test_membership_duration_empty(self):
        # Given
        data = []

        # When
        result = member_profile._membership_duration(data)

        # Then
        assert result == 0

    def test_process_personal_learning_plan(self):
        # TODO - lxml
        pass

    def test_process_role_data(self):
        # TODO - lxml
        pass

    def test_compile_ongoing_learning(self):
        # TODO - lxml
        pass


class TestScrapersMemberReduceDateList:
    def test_empty(self):
        # Given
        data = []

        # When
        result = member_profile._reduce_date_list(data)

        # Then
        assert result == []

    def test_single(self):
        # Given
        data = [(datetime.date(2015, 1, 1), datetime.date(2021, 1, 1))]

        # When
        result = member_profile._reduce_date_list(data)

        # Then
        assert result == data

    def test_outwith(self):
        # Given
        data = [
            (datetime.date(2016, 4, 5), datetime.date(2020, 2, 7)),
            (datetime.date(2015, 1, 1), datetime.date(2021, 1, 1)),
        ]

        # When
        result = member_profile._reduce_date_list(data)

        # Then
        assert result == [(datetime.date(2015, 1, 1), datetime.date(2021, 1, 1))]

    def test_starts_before_overlap(self):
        # Given
        data = [
            (datetime.date(2016, 4, 5), datetime.date(2020, 2, 7)),
            (datetime.date(2015, 1, 1), datetime.date(2018, 1, 1)),
        ]

        # When
        result = member_profile._reduce_date_list(data)

        # Then
        assert result == [(datetime.date(2015, 1, 1), datetime.date(2020, 2, 7))]

    def test_ends_after_overlap(self):
        # Given
        data = [
            (datetime.date(2016, 4, 5), datetime.date(2020, 2, 7)),
            (datetime.date(2017, 1, 1), datetime.date(2021, 1, 1)),
        ]

        # When
        result = member_profile._reduce_date_list(data)

        # Then
        assert result == [(datetime.date(2016, 4, 5), datetime.date(2021, 1, 1))]

    def test_within(self):
        # Given
        data = [
            (datetime.date(2016, 4, 5), datetime.date(2020, 2, 7)),
            (datetime.date(2017, 1, 1), datetime.date(2019, 1, 1)),
        ]

        # When
        result = member_profile._reduce_date_list(data)

        # Then
        assert result == [(datetime.date(2016, 4, 5), datetime.date(2020, 2, 7))]

    def test_adjacent(self):
        # Given
        data = [
            (datetime.date(2017, 1, 1), datetime.date(2017, 6, 30)),
            (datetime.date(2017, 7, 1), datetime.date(2017, 12, 31)),
        ]

        # When
        result = member_profile._reduce_date_list(data)

        # Then
        assert result == [(datetime.date(2017, 1, 1), datetime.date(2017, 12, 31))]

    def test_adjacent_inverse(self):
        # Given
        data = [
            (datetime.date(2016, 7, 1), datetime.date(2016, 12, 31)),
            (datetime.date(2017, 1, 1), datetime.date(2017, 6, 30)),
        ]

        # When
        result = member_profile._reduce_date_list(data)

        # Then
        assert result == [(datetime.date(2016, 7, 1), datetime.date(2017, 6, 30))]

    def test_disjoint(self):
        # Given
        data = [
            (datetime.date(2016, 7, 1), datetime.date(2016, 12, 31)),
            (datetime.date(2017, 7, 1), datetime.date(2017, 12, 31)),
        ]

        # When
        result = member_profile._reduce_date_list(data)

        # Then
        assert result == data

    def test_complex(self):
        # Given
        data = [
            (datetime.date(2016, 4, 5), datetime.date(2020, 2, 7)),
            (datetime.date(2015, 1, 1), datetime.date(2021, 1, 1)),
            (datetime.date(2017, 1, 1), datetime.date(2021, 6, 1)),
            (datetime.date(2014, 7, 22), datetime.date(2015, 6, 10)),
            (datetime.date(2015, 1, 1), datetime.date(2015, 1, 1)),
            (datetime.date(2004, 5, 13), datetime.date(2007, 11, 1)),
            (datetime.date(1997, 7, 2), datetime.date(2007, 11, 1)),
            (datetime.date(1993, 2, 18), datetime.date(1997, 7, 2)),
            (datetime.date(1983, 11, 17), datetime.date(1985, 3, 23)),
        ]

        # When
        result = member_profile._reduce_date_list(data)

        # Then
        assert result == [
            (datetime.date(1983, 11, 17), datetime.date(1985, 3, 23)),
            (datetime.date(1993, 2, 18), datetime.date(2007, 11, 1)),
            (datetime.date(2014, 7, 22), datetime.date(2021, 6, 1)),
        ]



import datetime

import httpx
import pydantic
import pytest

from compass.core import logon
import compass.core as ci
from compass.core.schemas.logon import CompassPropsConst
from compass.core.schemas.logon import CompassPropsConstSys
from compass.core.schemas.logon import CompassPropsCRUD
from compass.core.schemas.logon import CompassPropsMaster
from compass.core.schemas.logon import CompassPropsMasterConst
from compass.core.schemas.logon import CompassPropsMasterSSO
from compass.core.schemas.logon import CompassPropsMasterSys
from compass.core.schemas.logon import CompassPropsMasterUser
from compass.core.schemas.logon import CompassPropsNav
from compass.core.schemas.logon import CompassPropsPage
from compass.core.schemas.logon import CompassPropsUser
from compass.core.settings import Settings
from compass.core.util.client import Client

from tests.util.fake_compass import asp_net_id
from tests.util.fake_compass import id_map
from tests.util.fake_compass import session_id_default
from tests.util.fake_compass import username_map

base_domain = "127.0.0.1"
base_url = "http://127.0.0.1:4200"
jk_value = "9b65d68f4aca0138b5bae4492e7cdfae220a834e3e69731d996be1ddbb496d32fd29497d4f9729525c9fbc77666bb520ea214c0802ea22b958e6ae525224fd15"


class TestLogon:
    def test_login_init(self):
        # Given
        Settings.base_url = base_url
        client = Client()
        client.cookies.set("ASP.NET_SessionId", asp_net_id, domain=Settings.base_domain)
        user_props = {"cn": 10000000, "mrn": 9000000, "on": 10000001, "lvl": "ORG", "jk": jk_value}
        session_id = session_id_default
        compass_props = ci.CompassProps.parse_obj({"master": {"user": user_props, "sys": {"session_id": session_id}}})
        current_role = "Role", "Place"

        # When
        session = logon.Logon(client=client, compass_props=compass_props, current_role=current_role)

        # Then
        assert isinstance(session._client, Client)
        assert isinstance(session._client, httpx.Client)
        assert session.compass_props == compass_props
        assert session.current_role == current_role
        assert isinstance(session.hierarchy, ci.HierarchyLevel)
        assert session.hierarchy.unit_id == 10000001
        assert session.hierarchy.level == "Organisation"
        assert session.membership_number == 10000000
        assert session.role_number == 9000000
        assert session._jk == user_props["jk"]
        assert session._asp_net_id == asp_net_id
        assert session._session_id == session_id_default

    def test_login_init_no_unit_number(self):
        # Given
        Settings.base_url = base_url
        client = Client()
        compass_props = ci.CompassProps.parse_obj({"master": {"user": {"lvl": "ORG"}}})
        current_role = "Role", "Place"

        # Then
        with pytest.raises(ci.CompassError, match="Unit Number and Level must be specified!"):
            # When
            logon.Logon(client=client, compass_props=compass_props, current_role=current_role)

    def test_login_init_no_unit_level(self):
        # Given
        Settings.base_url = base_url
        client = Client()
        compass_props = ci.CompassProps.parse_obj({"master": {"user": {"on": 10000001}}})
        current_role = "Role", "Place"

        # Then
        with pytest.raises(ci.CompassError, match="Unit Number and Level must be specified!"):
            # When
            logon.Logon(client=client, compass_props=compass_props, current_role=current_role)

    def test_login_init_no_membership_number(self):
        # Given
        Settings.base_url = base_url
        client = Client()
        user_props = {"mrn": 9000000, "on": 10000001, "lvl": "ORG", "jk": jk_value}
        compass_props = ci.CompassProps.parse_obj({"master": {"user": user_props}})
        current_role = "Role", "Place"

        # Then
        with pytest.raises(ci.CompassError, match="User IDs must be specified!"):
            # When
            logon.Logon(client=client, compass_props=compass_props, current_role=current_role)

    def test_login_init_no_role_number(self):
        # Given
        Settings.base_url = base_url
        client = Client()
        user_props = {"cn": 10000000, "on": 10000001, "lvl": "ORG", "jk": jk_value}
        compass_props = ci.CompassProps.parse_obj({"master": {"user": user_props}})
        current_role = "Role", "Place"

        # Then
        with pytest.raises(ci.CompassError, match="User IDs must be specified!"):
            # When
            logon.Logon(client=client, compass_props=compass_props, current_role=current_role)

    def test_login_init_no_jk(self):
        # Given
        Settings.base_url = base_url
        client = Client()
        user_props = {"cn": 10000000, "mrn": 9000000, "on": 10000001, "lvl": "ORG"}
        compass_props = ci.CompassProps.parse_obj({"master": {"user": user_props}})
        current_role = "Role", "Place"

        # Then
        with pytest.raises(ci.CompassError, match="User IDs must be specified!"):
            # When
            logon.Logon(client=client, compass_props=compass_props, current_role=current_role)

    def test_login_init_no_session_id(self):
        # Given
        Settings.base_url = base_url
        client = Client()
        client.cookies.set("ASP.NET_SessionId", asp_net_id, domain=Settings.base_domain)
        user_props = {"cn": 10000000, "mrn": 9000000, "on": 10000001, "lvl": "ORG", "jk": jk_value}
        compass_props = ci.CompassProps.parse_obj({"master": {"user": user_props}})
        current_role = "Role", "Place"

        # Then
        with pytest.raises(ci.CompassError, match="ASP.NET ID must be specified!"):
            # When
            logon.Logon(client=client, compass_props=compass_props, current_role=current_role)

    def test_login_from_logon(self, server):
        # Given
        Settings.base_url = base_url
        credentials = "username", "password"

        # When
        session = logon.Logon.from_logon(credentials)

        # Then
        assert isinstance(session._asp_net_id, str)
        assert session._asp_net_id.islower()
        assert session._asp_net_id.isalnum()
        assert len(session._asp_net_id) == 24
        assert session._asp_net_id == session._client.cookies["ASP.NET_SessionId"]

        assert session._session_id == id_map[session._asp_net_id]
        assert session._client.headers["SID"] == id_map[session._asp_net_id]

        assert session._client.headers["Authorization"] == "10000000~9000000"

        assert session.current_role == ("Regional Administrator", "Wessex")
        assert session.hierarchy.unit_id == 10000001
        assert session.hierarchy.level == "Organisation"
        assert session.membership_number == 10000000
        assert session.role_number == 9000000
        assert session._jk == jk_value

    def test_login_from_logon_no_member_number(self, server):
        # Given
        Settings.base_url = base_url
        credentials = "no_role_number", "password"

        # Then
        with pytest.raises(ci.CompassError):
            # When
            logon.Logon.from_logon(credentials)

    @pytest.mark.skip("Changing roles is not implemented yet in fake_compass.")
    def test_login_from_logon_role_to_use(self, server):
        # TODO add change role functionality in fake_compass
        pass

    def test_login_from_session(self):
        # Given
        Settings.base_url = base_url
        user_props = {
            "cn": 10000000,
            "mrn": 9000000,
            "on": 10000001,
            "lvl": "ORG",
            "jk": jk_value,
        }
        session_id = session_id_default
        current_role = "Role", "Place"

        # When
        session = logon.Logon.from_session(asp_net_id, user_props, session_id, current_role)

        # Then
        assert session._client.cookies["ASP.NET_SessionId"] == asp_net_id
        assert session._asp_net_id == asp_net_id

        assert session._session_id == session_id
        assert session._client.headers["SID"] == session_id

        assert session._client.headers["Authorization"] == f'{user_props["cn"]}~{user_props["mrn"]}'

        assert session.current_role == current_role
        assert session.membership_number == user_props["cn"]
        assert session.role_number == user_props["mrn"]
        assert session._jk == user_props["jk"]

    def test_login_repr(self):
        # Given
        Settings.base_url = base_url
        client = Client()
        client.cookies.set("ASP.NET_SessionId", asp_net_id, domain=Settings.base_domain)
        user_props = {"cn": 10000000, "mrn": 9000000, "on": 10000001, "lvl": "ORG", "jk": jk_value}
        session_id = session_id_default
        compass_props = ci.CompassProps.parse_obj({"master": {"user": user_props, "sys": {"session_id": session_id}}})
        current_role = "Role", "Place"

        # When
        session = logon.Logon(client=client, compass_props=compass_props, current_role=current_role)

        # Then
        assert repr(session) == "<Logon> Compass ID: 10000000 (Role - Place)"

    def test_login_get_cookie(self, server):
        # Given
        Settings.base_url = base_url

        # When
        client = logon._create_session()

        # Then
        assert isinstance(client, httpx.Client)
        assert isinstance(client, Client)
        assert isinstance(client.cookies["ASP.NET_SessionId"], str)
        assert client.cookies["ASP.NET_SessionId"].islower()
        assert client.cookies["ASP.NET_SessionId"].isalnum()
        assert len(client.cookies["ASP.NET_SessionId"]) == 24

    def test_login_no_cookie(self, server):
        # Given
        Settings.base_url = f"{base_url}/_testing/no-cookie"

        # Then
        with pytest.raises(ci.CompassError, match="Could not create a session with Compass"):
            # When
            logon._create_session()

    # TODO this test is useless
    def test_login_post_credentials(self, server, monkeypatch: pytest.MonkeyPatch):
        # Given
        Settings.base_url = base_url
        client = Client()
        client.cookies.set("ASP.NET_SessionId", asp_net_id, domain=base_domain)

        # When
        monkeypatch.setattr(logon, "_check_login", lambda _client: (ci.CompassProps(), {}))
        _props, _roles = logon._logon_remote(client, ("username", "password"))

        # Then
        assert client.cookies["ASP.NET_SessionId"] == asp_net_id  # always need cookie

    # TODO this test is useless
    def test_login_post_incorrect_credentials(self, server, monkeypatch: pytest.MonkeyPatch):
        # Given
        Settings.base_url = base_url
        client = Client()
        client.cookies.set("ASP.NET_SessionId", asp_net_id, domain=base_domain)

        # When
        monkeypatch.setattr(logon, "_check_login", lambda _client: (ci.CompassProps(), {}))
        _props, _roles = logon._logon_remote(client, ("wrong", "credentials"))

        # Then
        assert client.cookies["ASP.NET_SessionId"] == asp_net_id  # always need cookie

    def test_login_check_login(self, server):
        # Given
        Settings.base_url = base_url
        client = Client()
        client.cookies.set("ASP.NET_SessionId", asp_net_id, domain=base_domain)
        id_map[asp_net_id] = session_id_default
        username_map[asp_net_id] = "username"

        # When
        props, roles = logon._check_login(client)

        # Then
        expected_props = ci.CompassProps(
            nav=CompassPropsNav(action="None", start_no=-1, start_page=3),
            page=CompassPropsPage(
                use_cn=10000000,
                hide_badges=True,
                croc="OK",
                hide_nominations=True,
                can_delete_ogl_hrs=False,
                fold_name="MP_",
            ),
            crud=CompassPropsCRUD(mdis="R", roles="R", pemd="R", mmmd="R", mvid="R", perm="R", trn="CRD"),
            user=CompassPropsUser(is_me=True),
            master=CompassPropsMaster(
                sso=CompassPropsMasterSSO(on=10000001),
                user=CompassPropsMasterUser(
                    cn=10000000,
                    mrn=9000000,
                    on=10000001,
                    lvl="ORG",
                    jk=jk_value,
                ),
                const=CompassPropsMasterConst(wales=10000007, scotland=10000005, over_seas=10000002, hq=10000001),
                sys=CompassPropsMasterSys(
                    session_id=session_id_default,
                    safe_json=True,
                    web_path=pydantic.HttpUrl(
                        "https://compass.scouts.org.uk/JSon.svc/",
                        scheme="https",
                        host="compass.scouts.org.uk",
                        tld="uk",
                        host_type="domain",
                        path="/JSon.svc/",
                    ),
                    text_size=1,
                    timout=1729000,
                    rest=True,
                    hard_time=datetime.time(20, 18, 28),
                    hard_expiry="6 hours",
                    timeout_extension=10,
                    ping=300000,
                ),
            ),
            const=CompassPropsConst(sys=CompassPropsConstSys(sto=5, sto_ask=0)),
        )

        expected_roles = {
            9000000: ("Regional Administrator", "Wessex"),
            6857721: ("TSA Council Member - Nominated Member (18-24)", "The Scout Association"),
        }

        assert client.cookies["ASP.NET_SessionId"] == asp_net_id  # always need cookie
        assert props == expected_props
        assert roles == expected_roles



import pytest

import compass.core as ci


class TestErrors:
    def test_compass_error(self):
        # Given
        data = "message"

        # Then
        with pytest.raises(ci.CompassError, match=data):
            # When
            raise ci.CompassError(data)

    def test_compass_authentication_error(self):
        # Given
        data = "message"

        # Then
        with pytest.raises(ci.CompassAuthenticationError, match=data):
            # When
            raise ci.CompassAuthenticationError(data)

    def test_compass_permission_error(self):
        # Given
        data = "message"

        # Then
        with pytest.raises(ci.CompassPermissionError, match=data):
            # When
            raise ci.CompassPermissionError(data)

    def test_compass_report_error(self):
        # Given
        data = "message"

        # Then
        with pytest.raises(ci.CompassReportError, match=data):
            # When
            raise ci.CompassReportError(data)

    def test_compass_report_permission_error(self):
        # Given
        data = "message"

        # Then
        with pytest.raises(ci.CompassReportPermissionError, match=data):
            # When
            raise ci.CompassReportPermissionError(data)

    def test_compass_authentication_error_inheritance(self):
        # Given
        data = "message"

        # When
        try:
            raise ci.CompassAuthenticationError(data)
        # Then
        except ci.CompassError as err:
            assert str(err) == data

    def test_compass_permission_error_inheritance(self):
        # Given
        data = "message"

        # When
        try:
            raise ci.CompassPermissionError(data)
        # Then
        except ci.CompassError as err:
            assert str(err) == data

    def test_compass_report_error_inheritance(self):
        # Given
        data = "message"

        # When
        try:
            raise ci.CompassReportError(data)
        # Then
        except ci.CompassError as err:
            assert str(err) == data

    def test_compass_report_permission_error_inheritance(self):
        # Given
        data = "message"

        # Dual inheritance:

        # When
        try:
            raise ci.CompassReportPermissionError(data)
        # Then
        except ci.CompassReportError as err:
            assert str(err) == data

        # When
        try:
            raise ci.CompassReportPermissionError(data)
        # Then
        except ci.CompassPermissionError as err:
            assert str(err) == data

        # When
        try:
            raise ci.CompassReportPermissionError(data)
        # Then
        except ci.CompassError as err:
            assert str(err) == data



"""Draft implementation of Compass -> Office 365 synchronisation."""

from __future__ import annotations

import datetime
import re
import secrets
import string
import time
from typing import Any, TYPE_CHECKING

import click
import httpx
import msal
import tomli

import compass.core as ci

if TYPE_CHECKING:
    from collections.abc import Set
    from io import TextIOWrapper

# ## ### LIBRARY SECTION ### ## #

# Globals
token_requested_time: float = 0
MEMBERS: dict[int, dict[str, Any]] = {}

# Constants
PWD_CHARS = string.ascii_letters + string.digits
WHITELISTED = re.compile("[^a-zA-Z-]+")
GRAPH_URL = "https://graph.microsoft.com/v1.0"
GRAPH_SCOPE = ["https://graph.microsoft.com/.default"]


def _get_member_ids() -> Set[int]:
    return MEMBERS.keys()


def _get_member_details(member_num: int) -> dict[str, Any]:
    return MEMBERS[member_num]


def _update_member_details(member_num: int, details: dict[str, Any]) -> None:
    try:
        MEMBERS[member_num] |= details
    # if member_num not in database, create entry with details dict
    except KeyError:
        MEMBERS[member_num] = details


def sync_compass_o365(
    api: ci.CompassInterface,
    domain: str,
    organisation_name: str,
    graph_auth: tuple[str, str, str],
    session: httpx.Client,
) -> None:
    # Fetch summary member report from Compass.
    member_ids_compass = api.hierarchy.unique_members(use_default=True)
    member_ids_local = _get_member_ids()

    # Gone Members - in database but missing from Compass
    old_member_ids = member_ids_local - member_ids_compass
    # New members - in Compass but missing from database
    new_member_ids = member_ids_compass - member_ids_local

    for member_num in old_member_ids:
        _update_member_details(member_num, {"in_compass": False, "last_update": datetime.date.today()})

    for member_num in new_member_ids:
        _add_member_details(api, member_num)

        _add_member_to_o365(member_num, domain, organisation_name, graph_auth, session)

    for member_num in member_ids_local:
        member_details = _get_member_details(member_num)
        if member_details["in_compass"] is False:
            print(f"Member {member_num} marked missing on {member_details['last_update']}")


def _add_member_details(api: ci.CompassInterface, member_num: int) -> None:
    personal_details = api.people.personal(member_num)
    roles_details = api.people.roles(member_num)

    _update_member_details(
        member_num,
        {
            "_id": member_num,
            "in_compass": True,
            "last_update": datetime.date.today(),
            "forenames": personal_details.forenames,
            "surname": personal_details.surname,
            # "full_name": personal_details.name,
            "known_as": personal_details.known_as,
            "display_name": f"{personal_details.known_as} {personal_details.surname}",
            # "address": personal_details.address,
            # "main_phone": personal_details.main_phone,
            "main_email": personal_details.main_email,
            # "join_date": personal_details.join_date,
            # "sex": personal_details.sex,
            "primary_role": roles_details.roles[roles_details.primary_role].role_title,
            # "primary_role_id": roles.primary_role,
            "primary_role_location": roles_details.roles[roles_details.primary_role].location_name,
            # "roles": roles.roles,
        },
    )


def _setup_session(graph_app_id: str, graph_authority: str, graph_app_secret: str, session: httpx.Client) -> None:
    global token_requested_time

    # Check the time since the header was last set, if older than 55 mins, reset the header to request a new token
    if time.time() - token_requested_time < (60 * 55):
        return
    print("Token expired - resetting")

    # acquire token
    app = msal.ConfidentialClientApplication(graph_app_id, authority=graph_authority, client_credential=graph_app_secret)
    token = app.acquire_token_silent(GRAPH_SCOPE, account=None)
    if not token:
        token = app.acquire_token_for_client(scopes=GRAPH_SCOPE)
        if not token or "error" in token:
            raise RuntimeError("Failed to get token from Azure AD")

    # Update headers and set update time
    session.headers.update({"Authorization": f"Bearer {token['access_token']}"})
    token_requested_time = time.time()


def _add_member_to_o365(
    membership_id: int,
    domain: str,
    organisation_name: str,
    graph_auth: tuple[str, str, str],
    session: httpx.Client,
) -> None:
    """Creates a new user on Office 365."""
    # Lookup the member from Compass
    try:
        member_obj = _get_member_details(membership_id)
    except KeyError:
        return None

    # Check an account hasn't already been created
    if "o365_profile" in member_obj:
        return None

    # Create the email address - keep only letters and hyphens
    firstname = WHITELISTED.sub("", member_obj["known_as"])
    surname = WHITELISTED.sub("", member_obj["surname"])
    email_nickname = f"{firstname}.{surname}".lower()
    user_principal_name = f"{email_nickname}@{domain}"

    # Generate a temp password
    tmp_password = "".join(secrets.choice(PWD_CHARS) for _ in range(20))

    create_obj = {
        # required
        "accountEnabled": True,
        "displayName": member_obj["display_name"],
        "mailNickname": email_nickname,
        "userPrincipalName": user_principal_name,
        "passwordProfile": {"forceChangePasswordNextSignIn": True, "password": tmp_password},
        # additional
        "givenName": member_obj["forenames"],
        "surname": member_obj["surname"],
        "employeeId": str(member_obj["_id"]),
        "usageLocation": "GB",
        "jobTitle": member_obj.get("o365_title_override", member_obj["primary_role"]),  # use title override if present
        "department": member_obj["primary_role_location"],
        "companyName": organisation_name,
        "otherMails": [member_obj["main_email"]],
    }

    graph_app_id, graph_authority, graph_app_secret = graph_auth

    # Create the user account
    _setup_session(graph_app_id, graph_authority, graph_app_secret, session)

    # result = session.post(f"{GRAPH_URL}/users", json=create_obj, headers={"Content-Type": "application/json"})
    result = session.post(f"{GRAPH_URL}/users", json=create_obj)
    if result.status_code != 201:
        raise RuntimeError(f"Failed to create Office 365 account with details: {create_obj}")

    o365_id = result.json()["id"]
    _update_member_details(
        membership_id, {"o365_profile": {"id": o365_id, "mail": user_principal_name}, "last_o365_sync": datetime.datetime.now()}
    )

    # assign license
    _setup_session(graph_app_id, graph_authority, graph_app_secret, session)

    # https://docs.microsoft.com/en-us/azure/active-directory/enterprise-users/licensing-service-plan-reference
    # Office 365 E1: 18181a46-0d4e-45cd-891e-60aabd171b4e
    # Office 365 E2: 6634e0ce-1a9f-428c-a498-f84ec7b8aa2e
    # Office 365 E3: 6fd2c87f-b296-42f0-b197-1e91e994b900
    # Office 365 E4: 1392051d-0cb9-4b7a-88d5-621fee5e8711
    # Office 365 E5: c7df2760-2c81-4ef7-b578-5b5392b571df
    # Microsoft 365 E3: 05e9a617-0261-4cee-bb44-138d3ef5d965
    # Microsoft 365 E5: 06ebc4ee-1bb5-47dd-8120-11324bc54e06
    # Enterprise Mobility & Security: efccb6f7-5641-4e0e-bd10-b4976e1bf68e

    o365_license = {"addLicenses": [{"skuId": "6634e0ce-1a9f-428c-a498-f84ec7b8aa2e"}], "removeLicenses": []}
    result = session.post(f"{GRAPH_URL}/users/{o365_id}/assignLicense", json=o365_license)
    if result.status_code != 200:
        raise RuntimeError(f"ERROR assigning license to user: {o365_id}")

    return None


# ## ### CLI SECTION ### ## #


def read_config_from_file(ctx: click.Context, _param: click.Parameter, file: TextIOWrapper | None) -> None:
    if file is None:
        return
    try:
        ctx.default_map = tomli.loads(file.read())
    except tomli.TOMLDecodeError:
        ctx.fail("Could not read config file!")


@click.command()
@click.option(
    "--config",
    type=click.File(mode="r", encoding="utf-8", lazy=False),
    callback=read_config_from_file,
    is_eager=True,
    expose_value=False,
    help="Read options from the specified toml-format configuration file. These are overridden by options passed on the"
    " command line.",
    show_default=True,
)
@click.option(
    "--compass-username",
    type=str,
    required=True,
    help="Username for logging in to Compass",
    envvar="COMPASS_USERNAME",
)
@click.option(
    "--compass-password",
    type=str,
    required=True,
    help="Password for logging in to Compass",
    envvar="COMPASS_PASSWORD",
)
@click.option(
    "--compass-role",
    type=str,
    help="Role for logging in to Compass. This is not required, but if given it must exactly match the name of the role"
    " as shown in your roles tab within Compass. By default, your primary role is used.",
    envvar="COMPASS_ROLE",
)
@click.option(
    "--compass-location",
    type=str,
    help="Location for logging in to Compass. This is not required, but if given it must exactly match the name of the"
    " location field for the corresponding role in the roles tab within Compass. Specifying the role location is only"
    " useful when you have multiple roles with the same role title.",
    envvar="COMPASS_LOCATION",
)
@click.option(
    "--tenant-id",
    type=str,
    required=True,
    help="Office 365 tenant ID. Should be of the form 'contoso.onmicrosoft.com'.",
    envvar="TENANT_ID",
)
@click.option(
    "--graph-app-id",
    type=str,
    required=True,
    help="Microsoft graph app ID.",
    envvar="GRAPH_APP_ID",
)
@click.option(
    "--graph-app-secret",
    type=str,
    required=True,
    help="Microsoft graph app secret.",
    envvar="GRAPH_APP_SECRET",
)
@click.option(
    "--o365-domain",
    type=str,
    required=True,
    help="Office 365 domain on which to assign user emails. Should be of the form 'example.org.uk'.",
    envvar="O365_DOMAIN",
)
@click.option(
    "--organisation-name",
    type=str,
    required=True,
    help="Your organisation's name, this will be set as users' company name. Should be of the form 'XYZ Scouts'.",
    envvar="ORG_NAME",
)
def cli(
    compass_username: str,
    compass_password: str,
    compass_role: str | None,
    compass_location: str | None,
    tenant_id: str,
    graph_app_id: str,
    graph_app_secret: str,
    o365_domain: str,
    organisation_name: str,
) -> None:
    api = ci.login(compass_username, compass_password, role=compass_role, location=compass_location)
    session = httpx.Client()

    # The actual function
    sync_compass_o365(
        api,
        o365_domain.lower(),
        organisation_name,
        (graph_app_id, f"https://login.microsoftonline.com/{tenant_id}", graph_app_secret),
        session,
    )


if __name__ == "__main__":
    cli()



import pytest

import compass.core as ci
from compass.core.util import client


class TestClient:
    def test_custom_error(self):
        # assume underlying client is tested by library -- only test what we add
        # Given
        session = client.Client()

        # Then
        with pytest.raises(ci.CompassNetworkError):
            # When
            session.get("https://httpbin.org/delay/5", timeout=0.5)



from urllib.parse import urlparse

import _strptime

from compass.core.settings import Settings


class TestSettings:
    def test_base_url(self):
        # Given
        data = Settings.base_url

        # When
        result = urlparse(data)

        # Then
        assert isinstance(data, str)
        assert bool(result.scheme) and bool(result.hostname)

    def test_date_format(self):
        # Given
        data = Settings.date_format

        # When
        result = _strptime._TimeRE_cache.pattern(data)  # pylint: disable=protected-access

        # Then
        assert isinstance(data, str)
        assert bool(result)

    def test_org_number(self):
        # Given When
        data = Settings.org_number

        # Then
        assert isinstance(data, int)
        assert 0 < data < 100_000_000  # IDs are between 1 and 99,999,999

    def test_wcf_json_endpoint(self):
        # Given When
        data = Settings.wcf_json_endpoint

        assert isinstance(data, str)
        assert data[0] == "/"
        assert ".svc" in data

    def test_web_service_path(self):
        # Given
        data = Settings.web_service_path

        # When
        result = urlparse(data)

        # Then
        assert isinstance(data, str)
        assert bool(result.scheme) and bool(result.hostname) and bool(result.path)



from __future__ import annotations

from typing import TYPE_CHECKING

import pydantic
import pytest

from compass.core.settings import Settings
from compass.core.util import context_managers

if TYPE_CHECKING:
    import pathlib


class TestContextManagers:
    def test_filesystem_guard_file(self, tmp_path: pathlib.Path):
        # Given some file (with text)
        filename = tmp_path / "existent.txt"
        test_text = "test-text"
        filename.write_text(test_text, encoding="utf-8")

        # When we read the file with filesystem_guard
        with context_managers.filesystem_guard("message (test_filesystem_guard_file)"):
            out = filename.read_text(encoding="utf-8")

        # Then check filesystem_guard hasn't mutated the text
        assert test_text == out

    def test_filesystem_guard_no_file(self, tmp_path: pathlib.Path, caplog: pytest.LogCaptureFixture):
        # Given some file (doesn't exist)
        filename = tmp_path / "non-existent.txt"

        # When we read the file with filesystem_guard
        with context_managers.filesystem_guard("message (test_filesystem_guard_no_file)"):
            filename.read_text(encoding="utf-8")

        # Then check filesystem_guard hasn't logged the error with the custom message
        assert "message (test_filesystem_guard_no_file)" in caplog.text

    def test_filesystem_guard_chained(self, tmp_path: pathlib.Path):
        # Given some file (with text)
        filename = tmp_path / "chained-with-statements.txt"
        test_text = "test-text"
        filename.write_text(test_text, encoding="utf-8")

        # When we read the file with filesystem_guard with chained `with` statements
        with context_managers.filesystem_guard("message (test_filesystem_guard_file)"), open(filename, "r", encoding="utf-8") as f:
            out = f.read()

        # Then check filesystem_guard hasn't mutated the text
        assert test_text == out

    def test_validation_errors_logging_raise(self, caplog: pytest.LogCaptureFixture):
        # Given
        Settings.validation_errors = True

        class TestModel(pydantic.BaseModel):
            val: int

        # Then
        with pytest.raises(pydantic.ValidationError, match="1 validation error"):
            # When
            with context_managers.validation_errors_logging(28_980, name="ID Number Type"):
                TestModel(val="not an integer")  # type: ignore[arg-type]

        assert caplog.messages[0] == "Parsing Error! ID Number Type: 28980"

    def test_validation_errors_logging_swallow(self, caplog: pytest.LogCaptureFixture):
        # Given
        Settings.validation_errors = False

        class TestModel(pydantic.BaseModel):
            val: int

        # When
        with context_managers.validation_errors_logging(28_980, name="ID Number Type"):
            TestModel(val="not an integer")  # type: ignore[arg-type]

        # Then
        assert caplog.messages[0] == "Parsing Error! ID Number Type: 28980"
        err_type, err_value, traceback = caplog.records[0].exc_info
        assert err_type is pydantic.ValidationError



import logging
import pathlib
import sys

from compass.core import logger


class TestLogger:
    def test_enable_debug_logging(self, tmp_path: pathlib.Path):
        # Given
        log_file = tmp_path / "test_log.log"

        # When
        logger.enable_debug_logging(log_file=log_file)

        # Then
        assert logger.logger.level == logging.DEBUG
        assert isinstance(logger.logger.handlers[0], logging.StreamHandler)
        assert isinstance(logger.logger.handlers[1], logging.FileHandler)
        assert logger.logger.handlers[0].stream == sys.stdout
        assert "test_log.log" in logger.logger.handlers[1].baseFilename



import pytest

from compass.core import logon
import compass.core as ci
from compass.core.settings import Settings
from compass.core.util.client import Client

from tests.util.fake_compass import asp_net_id
from tests.util.fake_compass import session_id_default

base_domain = "127.0.0.1"
base_url = "http://127.0.0.1:4200"
jk_value = "9b65d68f4aca0138b5bae4492e7cdfae220a834e3e69731d996be1ddbb496d32fd29497d4f9729525c9fbc77666bb520ea214c0802ea22b958e6ae525224fd15"


class TestInit:
    def test_compass_interface(self):
        # Given
        Settings.base_url = base_url
        client = Client()
        client.cookies.set("ASP.NET_SessionId", asp_net_id, domain=Settings.base_domain)
        user_props = {"cn": 10000000, "mrn": 9000000, "on": 10000001, "lvl": "ORG", "jk": jk_value}
        session_id = session_id_default
        compass_props = ci.CompassProps.parse_obj({"master": {"user": user_props, "sys": {"session_id": session_id}}})
        current_role = "Role", "Place"
        session = logon.Logon(client=client, compass_props=compass_props, current_role=current_role)

        # When
        api = ci.CompassInterface(session)

        # Then
        assert api.user is session
        assert isinstance(api.people, ci.People)
        assert isinstance(api.hierarchy, ci.Hierarchy)
        assert isinstance(api.reports, ci.Reports)

    def test_login_from_logon(self, server):
        # Given
        Settings.base_url = base_url
        username = "username"
        password = "password"  # nosec (false positive B105; it is a hardcoded password, but a fake one)

        # When
        api = ci.login(username, password)

        # Then
        assert isinstance(api, ci.CompassInterface)  # we test everything else elsewhere

    @pytest.mark.skip("Changing roles is not implemented yet in fake_compass.")
    def test_login_from_logon_with_role(self, server):
        # Given
        Settings.base_url = base_url
        username = "username"
        password = "password"  # nosec (false positive B105; it is a hardcoded password, but a fake one)

        # When
        api = ci.login(username, password, role="")

        # Then
        assert isinstance(api, ci.CompassInterface)  # we test everything else elsewhere

    @pytest.mark.skip("Changing roles is not implemented yet in fake_compass.")
    def test_login_from_logon_with_location(self, server):
        # Given
        Settings.base_url = base_url
        username = "username"
        password = "password"  # nosec (false positive B105; it is a hardcoded password, but a fake one)

        # When
        api = ci.login(username, password, location="")

        # Then
        assert isinstance(api, ci.CompassInterface)  # we test everything else elsewhere

    @pytest.mark.skip("Changing roles is not implemented yet in fake_compass.")
    def test_login_from_logon_with_role_and_location(self, server):
        # Given
        Settings.base_url = base_url
        username = "username"
        password = "password"  # nosec (false positive B105; it is a hardcoded password, but a fake one)

        # When
        api = ci.login(username, password, role="", location="")

        # Then
        assert isinstance(api, ci.CompassInterface)  # we test everything else elsewhere



from typing import Iterable, Literal

from compass.core.types.hierarchy import TYPES_UNIT_LEVELS

TYPES_REPORTS = Literal[
    "Appointments Report",
    "Member Directory Report",
    "18-25 Member Directory Report",
    "Permits Report",
    "Disclosure Report",
    "Training Report",
    "Awards Report",
    "Disclosure Management Report",
]
TYPES_FORMAT_CODE = Literal["CSV", "EXCEL", "XML"]
TYPES_FORMAT_CODES = Iterable[TYPES_FORMAT_CODE]
TYPES_EXPORTED_REPORTS = dict[TYPES_FORMAT_CODE, bytes]

# non-public
TYPES_REPORT_IDS_MAP = dict[TYPES_UNIT_LEVELS, int]



from __future__ import annotations

import compass.core as ci
from compass.core._scrapers.reports import export_report

__all__ = ("Reports",)
CSV: ci.TYPES_FORMAT_CODES = frozenset({"CSV"})


class Reports:
    """Interfaces with Compass's reporting engine.

    Exporting a report is of course surprisingly complicated. The process
    has four major steps, as follows:

    1. Get a token for generating the report from the Compass backend. This
        also validates that the report exists and that the user is
        authenticated to access it.
    2. Get the export URL from the report page, using the report token
        as a relative URL.
    3. Update report configuration data (sent as form data), and check that we
        are not in an error state. (Compass does not include all organisational
        units in reports by default, and we need to add in these missing/unset
        levels manually to export all data for a given unit and downwards).
    4. We export the prepared report in the specified format.

    Pitfalls to be aware of in this process include that:
    - Compass checks user-agent headers when updating form data
    - There is a ten (10) minute default soft-timeout, which may run out
        before a report download has finished
    - If a requested report is too large, Compass can simply give up, often
        with an `OutOfMemory` error or similar

    """

    def __init__(self, session: ci.Logon):
        """Constructor for Reports."""
        self.auth_ids = session.membership_number, session.role_number, session._jk
        self.client = session._client
        self.hierarchy_level = session.hierarchy.level

    def _get_report(self, report_type: ci.TYPES_REPORTS, formats: ci.TYPES_FORMAT_CODES = CSV) -> ci.TYPES_EXPORTED_REPORTS:
        """Exports report from Compass.

        Returns:
            Exported report content

        Raises:
            CompassReportError:
                - If the user passes an invalid report type
                - If Compass returns a JSON error
                - If there is an error updating the form data
            CompassReportPermissionError:
                If the user does not have permission to run the report
            CompassNetworkError:
                If there is an error in the transport layer, or if Compass
                reports a HTTP 5XX status code

        """
        return export_report(self.client, report_type, self.hierarchy_level, self.auth_ids, formats)

    def appointments(self, formats: ci.TYPES_FORMAT_CODES = CSV) -> ci.TYPES_EXPORTED_REPORTS:
        """Exports appointments reports from Compass.

        Returns:
            Exported report content

        Raises:
            CompassReportError:
                - If Compass returns a JSON error
                - If there is an error updating the form data
            CompassReportPermissionError:
                If the user does not have permission to run the report
            CompassNetworkError:
                If there is an error in the transport layer, or if Compass
                reports a HTTP 5XX status code

        """
        return self._get_report("Appointments Report", formats)

    def member_directory(self, formats: ci.TYPES_FORMAT_CODES = CSV) -> ci.TYPES_EXPORTED_REPORTS:
        """Exports member directory reports from Compass.

        Returns:
            Exported report content

        Raises:
            CompassReportError:
                - If Compass returns a JSON error
                - If there is an error updating the form data
            CompassReportPermissionError:
                If the user does not have permission to run the report
            CompassNetworkError:
                If there is an error in the transport layer, or if Compass
                reports a HTTP 5XX status code

        """
        return self._get_report("Member Directory Report", formats)

    def member_directory_18_25(self, formats: ci.TYPES_FORMAT_CODES = CSV) -> ci.TYPES_EXPORTED_REPORTS:
        """Exports member directory (18-25) reports from Compass.

        Returns:
            Exported report content

        Raises:
            CompassReportError:
                - If Compass returns a JSON error
                - If there is an error updating the form data
            CompassReportPermissionError:
                If the user does not have permission to run the report
            CompassNetworkError:
                If there is an error in the transport layer, or if Compass
                reports a HTTP 5XX status code

        """
        return self._get_report("18-25 Member Directory Report", formats)

    def permits(self, formats: ci.TYPES_FORMAT_CODES = CSV) -> ci.TYPES_EXPORTED_REPORTS:
        """Exports permits reports from Compass.

        Returns:
            Exported report content

        Raises:
            CompassReportError:
                - If Compass returns a JSON error
                - If there is an error updating the form data
            CompassReportPermissionError:
                If the user does not have permission to run the report
            CompassNetworkError:
                If there is an error in the transport layer, or if Compass
                reports a HTTP 5XX status code

        """
        return self._get_report("Permits Report", formats)

    def disclosure(self, formats: ci.TYPES_FORMAT_CODES = CSV) -> ci.TYPES_EXPORTED_REPORTS:
        """Exports disclosure reports from Compass.

        Returns:
            Exported report content

        Raises:
            CompassReportError:
                - If Compass returns a JSON error
                - If there is an error updating the form data
            CompassReportPermissionError:
                If the user does not have permission to run the report
            CompassNetworkError:
                If there is an error in the transport layer, or if Compass
                reports a HTTP 5XX status code

        """
        return self._get_report("Disclosure Report", formats)

    def training(self, formats: ci.TYPES_FORMAT_CODES = CSV) -> ci.TYPES_EXPORTED_REPORTS:
        """Exports training reports from Compass.

        Returns:
            Exported report content

        Raises:
            CompassReportError:
                - If Compass returns a JSON error
                - If there is an error updating the form data
            CompassReportPermissionError:
                If the user does not have permission to run the report
            CompassNetworkError:
                If there is an error in the transport layer, or if Compass
                reports a HTTP 5XX status code

        """
        return self._get_report("Training Report", formats)

    def awards(self, formats: ci.TYPES_FORMAT_CODES = CSV) -> ci.TYPES_EXPORTED_REPORTS:
        """Exports awards reports from Compass.

        Returns:
            Exported report content

        Raises:
            CompassReportError:
                - If Compass returns a JSON error
                - If there is an error updating the form data
            CompassReportPermissionError:
                If the user does not have permission to run the report
            CompassNetworkError:
                If there is an error in the transport layer, or if Compass
                reports a HTTP 5XX status code

        """
        return self._get_report("Awards Report", formats)

    def disclosure_management(self, formats: ci.TYPES_FORMAT_CODES = CSV) -> ci.TYPES_EXPORTED_REPORTS:
        """Exports disclosure management reports from Compass.

        Returns:
            Exported report content

        Raises:
            CompassReportError:
                - If Compass returns a JSON error
                - If there is an error updating the form data
            CompassReportPermissionError:
                If the user does not have permission to run the report
            CompassNetworkError:
                If there is an error in the transport layer, or if Compass
                reports a HTTP 5XX status code

        """
        return self._get_report("Disclosure Management Report", formats)



import hypothesis
import hypothesis.strategies as st

from compass.core.util import compass_helpers


class TestCompassHelpers:
    def test_hash_code(self):
        # Given
        data = "testing"

        # When
        result = compass_helpers.hash_code(data)

        # Then
        assert isinstance(result, int)
        assert result == -1422446064

    @hypothesis.given(st.text())
    def test_hash_code_range(self, data):
        # Given data from hypothesis, When
        result = compass_helpers.hash_code(data)

        # Then
        assert isinstance(result, int)
        assert -2_147_483_648 <= result <= 2_147_483_647  # -2**31 to 2**31-1

    @hypothesis.given(st.dictionaries(st.text(), st.none() | st.booleans() | st.floats() | st.integers() | st.text()), st.none())
    @hypothesis.example(
        data={"a": 1, "b": 2, "c": 3},
        expected=[{"Key": "a", "Value": "1"}, {"Key": "b", "Value": "2"}, {"Key": "c", "Value": "3"}],
    )
    def test_compass_restify(self, data, expected):
        # Given data from hypothesis
        size = len(data)

        # When
        result = compass_helpers.compass_restify(data)

        # Then
        assert isinstance(result, list)
        assert len(result) == size
        for pair in result:
            assert isinstance(pair, dict)
            assert len(pair) == 2
            assert pair.keys() == {"Key", "Value"}
            for val in pair.values():
                assert isinstance(val, str)
        if expected is not None:
            assert result == expected



from typing import Any

import httpx

import compass.core as ci
from compass.core.settings import Settings


class Client(httpx.Client):
    """Counts the number of requests sent."""

    def request(self, *args: Any, **kwargs: Any) -> httpx.Response:
        Settings.total_requests += 1
        try:
            return super().request(*args, **kwargs)
        except httpx.RequestError as err:
            raise ci.CompassNetworkError("Network Error!") from err



from __future__ import annotations

import re
import time
from typing import cast, TYPE_CHECKING

import httpx
from lxml import html

import compass.core as ci
from compass.core.logger import logger
from compass.core.settings import Settings
from compass.core.util import auth_header

if TYPE_CHECKING:
    from compass.core.types.reports import TYPES_REPORT_IDS_MAP
    from compass.core.util.auth_header import TYPE_AUTH_IDS
    from compass.core.util.client import Client

_report_ids_appointments: TYPES_REPORT_IDS_MAP = {
    "Group": 59,
    "District": 50,
    "County": 48,
    "Region": 52,
}
_report_ids_member_directory: TYPES_REPORT_IDS_MAP = {
    "District": 51,
    "County": 49,
    "Region": 37,
}
_report_ids_18_25_member_directory: TYPES_REPORT_IDS_MAP = {
    "County": 53,
}
_report_ids_permits: TYPES_REPORT_IDS_MAP = {
    "District": 70,
    "County": 69,
    "Region": 72,
}
_report_ids_disclosure: TYPES_REPORT_IDS_MAP = {
    "District": 78,
    "County": 77,
    "Region": 76,
}
_report_ids_training: TYPES_REPORT_IDS_MAP = {
    "District": 79,
    "County": 80,
    "Region": 84,
}
_report_ids_awards: TYPES_REPORT_IDS_MAP = {
    "District": 94,
    "County": 95,
}
_report_ids_disclosure_management: TYPES_REPORT_IDS_MAP = {
    "District": 102,
    "County": 101,
    "Region": 100,
}
_report_ids: dict[ci.TYPES_REPORTS, TYPES_REPORT_IDS_MAP] = {
    "Appointments Report": _report_ids_appointments,
    "Member Directory Report": _report_ids_member_directory,
    "18-25 Member Directory Report": _report_ids_18_25_member_directory,
    "Permits Report": _report_ids_permits,
    "Disclosure Report": _report_ids_disclosure,
    "Training Report": _report_ids_training,
    "Awards Report": _report_ids_awards,
    "Disclosure Management Report": _report_ids_disclosure_management,
}
_ADDITIONAL_FORM_DATA = {
    "ReportViewer1$ctl10": "ltr",
    "ReportViewer1$ctl11": "standards",
    "ReportViewer1$ctl05$ctl00$CurrentPage": "1",
    "ReportViewer1$ctl09$VisibilityState$ctl00": "ReportPage",
    "__EVENTTARGET": "ReportViewer1$ctl04$ctl07",
    "__EVENTARGUMENT": "",
    "__LASTFOCUS": "",
}


def export_report(
    client: Client,
    report_type: ci.TYPES_REPORTS,
    hierarchy_level: ci.TYPES_HIERARCHY_LEVELS,
    auth_ids: TYPE_AUTH_IDS,
    formats: ci.TYPES_FORMAT_CODES,
) -> ci.TYPES_EXPORTED_REPORTS:
    """Exports report as CSV from Compass.

    See `Reports.get_report` for an overview of the export process

    Returns:
        Report output, as a bytes-encoded object.

    Raises:
        CompassReportError:
            - If the user passes an invalid report type
            - If Compass returns a JSON error
            - If there is an error updating the form data
        CompassReportPermissionError:
            If the user does not have permission to run the report
        CompassNetworkError:
            If there is an error in the transport layer, or if Compass
            reports a HTTP 5XX status code

    """
    report_number = _report_number(report_type, hierarchy_level)
    logger.info(f"Generating {report_type.lower()}")
    report_page, run_report_url = _initialise_report(client, auth_ids, report_number)

    # Training Reports -- for regional reports, we want to run each county individually, as compass can't handle bigger
    if report_number == 84:
        return _training_report_region(client, report_page, run_report_url)

    _update_form_data(client, report_page, run_report_url, report_number)
    return _export_report(client, report_page, formats)


def _report_number(report_type: ci.TYPES_REPORTS, hierarchy_level: ci.TYPES_HIERARCHY_LEVELS) -> int:
    try:
        report_level_map = _report_ids[report_type]
    except KeyError:
        raise ci.CompassReportError(f"{report_type} is not a valid report type. Valid report types are {[*_report_ids]}") from None
    try:
        return report_level_map[cast(ci.TYPES_UNIT_LEVELS, hierarchy_level)]
    except KeyError:
        raise ci.CompassReportError(f"Requested report does not exist for hierarchy level: {hierarchy_level}.") from None


def _initialise_report(client: Client, auth_ids: TYPE_AUTH_IDS, report_number: int) -> tuple[str, str]:
    # Get token for report type & role running said report:
    run_report_url = _get_report_token(client, auth_ids, report_number)

    # Get initial reports page, for export URL and config:
    report_page = client.get(run_report_url).content

    return report_page.decode("utf-8"), run_report_url


def _get_report_token(client: Client, auth_ids: TYPE_AUTH_IDS, report_number: int) -> str:
    logger.debug("Getting report token")
    response = auth_header.auth_header_get(
        auth_ids,
        client,
        f"{Settings.web_service_path}/ReportToken",
        params={
            "pReportNumber": str(report_number),
            "pMemberRoleNumber": str(auth_ids[1]),  # auth IDs are membership number, role number, 'jk'
        },
    )
    _error_status(response)

    report_token_uri = response.json().get("d", "")
    if report_token_uri not in {"-1", "-2", "-3", "-4"}:
        return f"{Settings.base_url}/{report_token_uri}"
    if report_token_uri in {"-2", "-3"}:
        raise ci.CompassReportError("Report aborted: Report No Longer Available")
    if report_token_uri == "-4":  # nosec (false positive B105; not a hardcoded passwordstring)
        raise ci.CompassReportPermissionError("Report aborted: USER DOES NOT HAVE PERMISSION")
    raise ci.CompassReportError("Report aborted")


def _training_report_region(client: Client, report_page: str, run_report_url: str) -> ci.TYPES_EXPORTED_REPORTS:
    # pylint: disable=too-many-locals
    tree, form_data = _extract_form_data(report_page)
    numbered_counties = _parse_drop_down_list(tree, "ReportViewer1_ctl04_ctl05_divDropDown")

    county_reports = []
    for level, county in numbered_counties.items():
        logger.info(f"Starting training report export for {county}")
        # Update report for current county
        cty_form_data = form_data | {
            "ReportViewer1$ctl04$ctl05$txtValue": county,
            "ReportViewer1$ctl04$ctl05$divDropDown$ctl01$HiddenIndices": level,
            "ReportViewer1$ctl04$ctl07$txtValue": "Selected Level Only",  # set the defaults
        }
        cty_form_data |= _ADDITIONAL_FORM_DATA | {"__EVENTTARGET": "ReportViewer1$ctl04$ctl05"}
        county_report_page = client.post(
            run_report_url,
            data=cty_form_data,
            headers={"User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64)"},
        )

        if county != "Region Level":
            # get current county's districts
            cty_tree, cty_form_data = _extract_form_data(county_report_page.content.decode("utf-8"))
            table = cty_tree.get_element_by_id("ReportViewer1_ctl04_ctl07_divDropDown")[0]
            numbered_districts = {str(i): row[0][0][1].text.replace("\xa0", " ") for i, row in enumerate(table[1:])}

            # Update report for current county's districts
            cty_form_data["ReportViewer1$ctl04$ctl07$txtValue"] = ", ".join(numbered_districts.values())
            cty_form_data["ReportViewer1$ctl04$ctl07$divDropDown$ctl01$HiddenIndices"] = ",".join(numbered_districts.keys())
            _update_report_with_form_data(client, run_report_url, cty_form_data)

        try:
            # hardcode CSV, there's enough complexity as it is
            county_reports.append(_export_report(client, report_page, {"CSV"})["CSV"])
        except ci.CompassNetworkError:
            logger.error(f"Failed to export training report for {county.title()} county!")

    # Normalise the reports. This assumes a lot about the structure & prefix of the reports...
    prefix = b"\xef\xbb\xbfMember_Training_Report\r\nDate : " + time.strftime("%d/%m/%Y").encode() + b"\r\n\r\n"
    out = prefix
    for report in county_reports:
        if report.startswith(prefix):
            out += report[48:]  # 48 == len(prefix)
        else:
            logger.error(f"Training report does not start with common prefix! First 50 chars are {repr(report[:50])}")

    return {"CSV": out}


def _update_form_data(client: Client, report_page: str, run_report: str, report_number: int) -> None:
    """Update form data & set location selection."""
    # TODO add method to choose between exporting all data and just top-level
    tree, form_data = _extract_form_data(report_page)

    # Appointments Reports
    # TODO county training should probably be added to this list
    if report_number in {48, 52}:  # County, Region
        form_data = _form_data_appointments(form_data, tree)

    _update_report_with_form_data(client, run_report, form_data)


def _extract_form_data(report_page: str) -> tuple[html.HtmlElement, dict[str, str]]:
    tree = html.fromstring(report_page)

    # get relevant form data. TODO - do we need all items or just __VIEWSTATE?
    # form_data = {"__VIEWSTATE": next((el.value for el in tree.forms[0].iter("input") if el.name == "__VIEWSTATE"), "")}
    form_data = {el.name: el.value for el in tree.forms[0].inputs if el.get("type") not in {"checkbox", "image"}}

    return tree, form_data


def _update_report_with_form_data(client: Client, run_report_url: str, form_data: dict[str, str]) -> None:
    # Compass does user-agent sniffing in reports!!! This does seem to be the
    # only place that *requires* a Mozilla/5 type UA.
    # Including the MicrosoftAjax pair lets us check errors quickly. In reality
    # we don't care about the output of this POST, just that it doesn't fail.
    updated_report_page = client.post(
        run_report_url,
        data=form_data,
        headers={"User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64)", "X-MicrosoftAjax": "Delta=true"},
    )

    # Check error state
    _error_status(updated_report_page, msg="Updating report locations failed!")
    if "compass.scouts.org.uk%2fError.aspx|" in updated_report_page.text:
        raise ci.CompassReportError("Compass Error!")


def _form_data_appointments(form_data: dict[str, str], tree: html.HtmlElement) -> dict[str, str]:
    """Select all units/locations."""
    # report level - 1 (e.g. county -> district)
    numbered_levels_children = _parse_drop_down_list(tree, "ReportViewer1_ctl04_ctl05_divDropDown")
    form_data["ReportViewer1$ctl04$ctl05$txtValue"] = ", ".join(numbered_levels_children.values())
    form_data["ReportViewer1$ctl04$ctl05$divDropDown$ctl01$HiddenIndices"] = ",".join(numbered_levels_children.keys())

    # report level - 2 (e.g. county -> group)
    numbered_levels_grandchildren = _parse_drop_down_list(tree, "ReportViewer1_ctl04_ctl07_divDropDown")
    form_data["ReportViewer1$ctl04$ctl07$txtValue"] = ", ".join(numbered_levels_grandchildren.values())
    form_data["ReportViewer1$ctl04$ctl07$divDropDown$ctl01$HiddenIndices"] = ",".join(numbered_levels_grandchildren.keys())

    # TODO the additional form data may not be needed. Test.
    return form_data | _ADDITIONAL_FORM_DATA | {"__ASYNCPOST": "true"}


def _parse_drop_down_list(tree: html.HtmlElement, element_id: str, /) -> dict[str, str]:
    table = tree.get_element_by_id(element_id)[0][0]
    return {str(i): row[0][0][0][1].text.replace("\xa0", " ") for i, row in enumerate(table[1:])}


def _export_report(client: Client, report_page: str, formats: ci.TYPES_FORMAT_CODES) -> ci.TYPES_EXPORTED_REPORTS:
    # Get report export URL:
    export_url_base = _extract_report_export_url_base(report_page)

    # Exported reports storage
    exported = {}
    for format_code in set(formats):
        # Download report:
        logger.info(f"Exporting report to {format_code.lower()}")
        start = time.time()
        export_code = "EXCELOPENXML" if format_code == "EXCEL" else format_code  # correct excel exports
        exported[format_code] = client.get(export_url_base + export_code, timeout=300).content  # excel takes **ages**
        logger.debug(f"Downloading took {time.time() - start:.2f}s")

    # start = time.time()
    # TODO TRAINING REPORT ETC.
    # # TODO REPORT BODY HAS KEEP ALIVE URL KeepAliveUrl
    # p = PeriodicTimer(15, lambda: self.report_keep_alive(self.session, report_page.text))
    # self.session.sto_thread.start()
    # p.start()
    # # ska_url = _report_keep_alive(self.session, report_page.text)
    # try:
    #     _download_report(self.session, f"{Settings.base_url}/{export_url_path}", export_url_params, filename, )  # ska_url
    # except (ConnectionResetError, httpx.ConnectError):
    #     logger.info(f"Stopped at {datetime.datetime.now()}")
    #     p.cancel()
    #     self.session.sto_thread.cancel()
    #     raise
    # logger.debug(f"Exporting took {time.time() - start}s")

    return exported


def _extract_report_export_url_base(report_page: str) -> str:
    start = report_page.index("ExportUrlBase")
    cut = report_page[start:].removeprefix('ExportUrlBase":"')
    end = cut.index('"')
    full_url = cut[:end].encode().decode("unicode-escape")
    return f"{Settings.base_url}/{full_url}"


def _error_status(response: httpx.Response, /, msg: str = "Request to Compass failed!") -> None:
    try:
        response.raise_for_status()
    except httpx.HTTPError as err:
        raise ci.CompassNetworkError(msg) from err


def _report_keep_alive(client: Client, report_page: str) -> str:
    logger.info(f"Extending Report Session {time.strftime('%Y-%m-%d %H-%M-%S')}")
    keep_alive_encoded = re.search(r'"KeepAliveUrl":"(.*?)"', report_page).group(1)  # type: ignore[union-attr]
    keep_alive = keep_alive_encoded.encode().decode("unicode-escape")
    response = client.post(f"{Settings.base_url}{keep_alive}")  # NoQA: F841 (unused variable)

    return keep_alive  # response



from collections.abc import Iterator
import contextlib
import threading
import time

import uvicorn


# https://github.com/encode/uvicorn/issues/742#issuecomment-674411676
# TODO use https://github.com/encode/uvicorn/pull/917
class Server(uvicorn.Server):
    def install_signal_handlers(self):
        pass

    @contextlib.contextmanager
    def run_in_thread(self) -> Iterator[None]:
        thread = threading.Thread(target=self.run)
        thread.start()
        try:
            while not self.started:
                time.sleep(1e-3)
            yield
        finally:
            self.should_exit = True
            thread.join()



from __future__ import annotations

import datetime
from typing import Literal, Optional, Union
import warnings

import phonenumbers
import pydantic

from compass.core.types.member import TYPES_AWARD_TYPE
from compass.core.types.member import TYPES_DISCLOSURE_PROVIDERS
from compass.core.types.member import TYPES_DISCLOSURE_STATUSES
from compass.core.types.member import TYPES_DISCLOSURES_APPOINTMENT
from compass.core.types.member import TYPES_DISCLOSURES_COUNTRIES
from compass.core.types.member import TYPES_ETHNICITY
from compass.core.types.member import TYPES_LEARNING_METHOD
from compass.core.types.member import TYPES_OCCUPATION
from compass.core.types.member import TYPES_PERMIT
from compass.core.types.member import TYPES_PERMIT_TYPE
from compass.core.types.member import TYPES_REFERENCES
from compass.core.types.member import TYPES_RELIGION
from compass.core.types.member import TYPES_ROLE_CLASS
from compass.core.types.member import TYPES_ROLE_STATUS
from compass.core.types.member import TYPES_SEX


class MemberBase(pydantic.BaseModel):
    membership_number: int


class MemberRoleBase(pydantic.BaseModel):
    # Role details
    role_number: int
    role_title: str  # role_name
    role_start: datetime.date
    role_status: TYPES_ROLE_STATUS
    review_date: Optional[datetime.date] = None


class MemberDetailsAddress(pydantic.BaseModel):
    unparsed_address: Optional[str] = None  # this is the unmodified address string
    country: Optional[str] = None
    postcode: Optional[str] = None
    county: Optional[str] = None
    town: Optional[str] = None
    street: Optional[str] = None


# Personal Details Tab
class MemberDetails(MemberBase):
    # Name
    name: str
    known_as: str
    forenames: Optional[str] = None
    surname: Optional[str] = None

    # Core personal information
    birth_date: Optional[datetime.date] = None
    sex: Optional[TYPES_SEX] = None
    nationality: Optional[str] = None  # literal? Big list...!
    ethnicity: Optional[TYPES_ETHNICITY] = None
    religion: Optional[TYPES_RELIGION] = None
    religion_detail: Optional[str] = None
    occupation: Optional[TYPES_OCCUPATION] = None
    occupation_detail: Optional[str] = None
    join_date: Optional[datetime.date] = None

    # Contact Details
    main_phone: Optional[str] = None
    main_email: Optional[pydantic.EmailStr] = None
    address: MemberDetailsAddress = MemberDetailsAddress()

    # Additional / miscellaneous details (maybe add keyed models ref #86)
    disabilities: Optional[dict[str, str]] = None
    qualifications: Optional[dict[str, str]] = None
    hobbies: Optional[dict[str, str]] = None

    # pylint doesn't support custom decorators. https://github.com/PyCQA/pylint/issues/1694
    # pylint: disable=no-self-use
    @pydantic.validator("main_phone")
    def check_phone_number(cls, v: Optional[str], values: dict[str, object]) -> Optional[str]:
        if v is None or not v or v == "0" or len(v) < 2:
            return None  # checks for v in (None, "", "0", .{1,2})

        try:
            n = phonenumbers.parse(v, "GB")
            if phonenumbers.is_valid_number(n):
                if n.country_code == 44:
                    return f"{phonenumbers.format_number(n, phonenumbers.PhoneNumberFormat.NATIONAL)}"
                return f"{phonenumbers.format_number(n, phonenumbers.PhoneNumberFormat.INTERNATIONAL)}"
        except phonenumbers.NumberParseException:
            pass

        membership_number = values["membership_number"]
        warnings.warn(f"Member No {membership_number}: phone number {v} is not valid!", RuntimeWarning)
        return None  # mypy being silly


# Roles Tab (Main List - item)
class MemberRoleCore(MemberBase, MemberRoleBase):
    # Role details
    role_end: Optional[datetime.date] = None
    role_class: TYPES_ROLE_CLASS
    role_type: Optional[str] = None
    # Full list of role types runs to 458, probably too many for a Literal type
    # xref: https://compass.scouts.org.uk/Roles.aspx?closed=Y

    # Location details
    location_id: Optional[int] = None
    location_name: Optional[str] = None

    # Is the role details popup accessible?
    can_view_details: bool


# Roles Tab (Main List - collection)
class MemberRolesCollection(pydantic.BaseModel):
    roles: dict[int, MemberRoleCore]
    membership_duration: float  # Membership duration in qualifying roles, in years
    primary_role: Optional[int]  # Primary role number. None if no primary role found


# Roles Tab (Role Detail Popup - Main)
class MemberRoleDetail(MemberBase, MemberRoleBase):
    organisation_level: str
    birth_date: Optional[datetime.date]
    name: str

    # Approval Process
    line_manager_number: Optional[int] = None
    line_manager: Optional[str] = None

    # Approval information
    ce_check_valid: bool
    ce_check_date: Optional[datetime.date]  # Optional for Closed roles - e.g. #499, role closed 1976, or if Pending
    disclosure_check: Union[TYPES_DISCLOSURES_APPOINTMENT, None, pydantic.constr(regex=r"^Issued : \d\d [A-Z][a-z]+ \d\d\d\d$")]  # type: ignore[valid-type]  # NoQA: F722
    disclosure_date: Optional[datetime.date]
    references: Optional[TYPES_REFERENCES] = None
    appointment_panel_approval: Optional[Literal["NC", "NR", "S", "U"]] = None
    commissioner_approval: Optional[Literal["NC", "NR", "RR", "S", "U"]] = None
    committee_approval: Optional[Literal["NC", "NR", "S", "U"]] = None


# Roles Tab (Role Detail Popup - Hierarchy)
class MemberRoleHierarchy(pydantic.BaseModel):
    # Role location
    organisation: str
    country: Optional[str] = None
    region: Optional[str] = None
    county: Optional[str] = None
    district: Optional[str] = None
    group: Optional[str] = None
    section: Optional[str] = None


# Roles Tab (Role Detail Popup - Getting Started)
class MemberRoleGettingStartedModule(pydantic.BaseModel):
    validated: Optional[datetime.date] = None
    validated_by: Optional[str] = None


# Roles Tab (Role Detail Popup - Getting Started)
class MemberRoleGettingStarted(pydantic.BaseModel):
    """Getting Started Training."""

    module_01: Optional[MemberRoleGettingStartedModule] = None
    trustee_intro: Optional[MemberRoleGettingStartedModule] = None
    module_02: Optional[MemberRoleGettingStartedModule] = None
    module_03: Optional[MemberRoleGettingStartedModule] = None
    module_04: Optional[MemberRoleGettingStartedModule] = None
    gdpr: Optional[MemberRoleGettingStartedModule] = None
    safety: Optional[MemberRoleGettingStartedModule]
    safeguarding: Optional[MemberRoleGettingStartedModule]


# Roles Tab (Role Detail Popup - All)
class MemberRolePopup(pydantic.BaseModel):
    hierarchy: MemberRoleHierarchy
    details: MemberRoleDetail
    getting_started: MemberRoleGettingStarted


# Training Tab (Role)
class MemberTrainingRole(MemberRoleBase):
    # Location details
    location: Optional[str] = None

    # Training Advisor details
    ta_data: Optional[str]
    ta_number: Optional[int]
    ta_name: Optional[str]

    # Completion details
    completion: Optional[str]
    completion_type: Optional[Literal["PLP", "Wood Badge"]]
    completion_date: Optional[datetime.date]
    wood_badge_number: Optional[int]


# Training Tab (PLP)
class MemberTrainingPLP(pydantic.BaseModel):
    pk: int
    module_id: int
    code: str  # We could do literals for module name/code but this would be ~150 values
    name: str

    # Learning details
    learning_required: Optional[bool]
    learning_method: Optional[TYPES_LEARNING_METHOD]
    learning_completed: Optional[datetime.date]
    learning_date: Optional[datetime.date]

    # Validation details
    validated_membership_number: Optional[int]
    validated_name: Optional[str]
    validated_date: Optional[datetime.date]


# Training Tab (OGL - item)
class MemberMOGL(pydantic.BaseModel):
    completed_date: Optional[datetime.date]
    renewal_date: Optional[datetime.date]  # GDPR has no renewal date...??


# Training Tab (OGL - all)
class MemberMandatoryTraining(pydantic.BaseModel):
    safety: MemberMOGL
    safeguarding: MemberMOGL
    first_aid: MemberMOGL
    gdpr: MemberMOGL


# Training Tab
class MemberTrainingTab(pydantic.BaseModel):
    roles: dict[int, MemberTrainingRole]
    plps: dict[int, list[MemberTrainingPLP]]
    mandatory: MemberMandatoryTraining


# Permits Tab
class MemberPermit(MemberBase):
    permit_type: TYPES_PERMIT
    category: str
    type: TYPES_PERMIT_TYPE
    restrictions: str
    expires: Optional[datetime.date]
    status: Literal["PERM_OK", "PERM_EXP", "PERM_REV"]


class MemberAward(MemberBase):
    type: TYPES_AWARD_TYPE
    location: Optional[str]
    date: datetime.date


_DISCLOSURE_NUM_SCOTLAND = Union[pydantic.constr(regex=r"^\d{7}R$"), Literal["Non-OSCR Reg"]]  # type: ignore[misc]  # NoQA: F722
_DISCLOSURE_NUM_NI = pydantic.constr(regex=r"^PECS(VC)?\d{6}(CM|BM|RL)$")  # PECS was Pre-Employment Consultancy Service


class MemberDisclosure(MemberBase):
    country: TYPES_DISCLOSURES_COUNTRIES
    provider: TYPES_DISCLOSURE_PROVIDERS
    # enhanced with barring is almost always the case.
    type: Literal["Enhanced with Barring", "Enhanced", "Local Check", "Adult Enhanced with Barring", "Unknown Disclosure Type"]
    # If Application Withdrawn, no disclosure number. If Scottish in the early 2000s, 7 digits ending with an R
    number: Union[int, None, _DISCLOSURE_NUM_SCOTLAND, _DISCLOSURE_NUM_NI]  # type: ignore[valid-type]  # NoQA: F722
    issuer: Optional[TYPES_DISCLOSURE_PROVIDERS]
    issue_date: Optional[datetime.date]  # If Application Withdrawn, maybe no issue date
    status: TYPES_DISCLOSURE_STATUSES
    expiry_date: Optional[datetime.date]  # If Application Withdrawn, no expiry date
